# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils to read envlogger data into a dataset that is compatible with RLDS."""

from typing import Any, Callable, Dict, Generator, List, Optional, Sequence, Tuple, Union

import numpy as np
from tensorflow_datasets.core.utils import tree_utils


def _get_episode_metadata(episode: Sequence[Any]) -> Dict[str, Any]:
  """Extracts the episode metadata and returns it as a dictionary.

  This is only used when the episode metadata is stored in the first step.

  Args:
    episode: sequence of steps (envlogger.StepData).

  Returns:
    Dictionary containing the episode metadata
  """

  first_step = episode[0]
  if not first_step.timestep.first():
    raise ValueError('The first step in this episode is not first')
  custom_data = first_step.custom_data
  return custom_data.get('episode_metadata', {}) if custom_data else {}


def _get_step_metadata(step: Any) -> Dict[str, Any]:
  """Extracts the custom metadata from the step.

  Note that it may return the dictionary contained in the step, so further
  modifications will affect the step itself.

  Args:
    step: step data (envlogger.StepData).

  Returns:
    Dictionary containing the step metadata.
  """
  if not isinstance(step.custom_data, dict):
    # We ignore step metadata if step.custom_data is not a dictionary
    return {}
  if 'episode_metadata' in step.custom_data:
    # Some datasets used an old version of EnvLogger that stores episode
    # metadata in the first step.
    return {
        k: v for k, v in step.custom_data.items() if k != 'episode_metadata'
    }
  else:
    return step.custom_data


def get_episode_dict(
    episode: Sequence[Any],
    episode_metadata: Dict[str, Any],
    episode_metadata_fn: Callable[
        [Dict[str, Any]], Optional[Dict[str, Any]]
    ] = lambda x: x,
    step_fn: Callable[[Dict[str, Any]], Dict[str, Any]] = lambda x: x,
) -> Optional[Dict[str, Any]]:
  """Obtains an RLDS episode from an Envlogger episode.

  Args:
    episode: sequence of envlogger.StepData.
    episode_metadata: metadata as obtained from the envlogger reader.
    episode_metadata_fn: function to process the episode metadata. If it returns
      None, the steps are not read and this function returns None.
    step_fn: function that processes each RLDS step once it's built from the raw
      data.

  Returns:
     Episode (or None if episode_metadata_fn returns None).
  """
  if episode_metadata:
    episode_dict = episode_metadata
  else:
    episode_dict = _get_episode_metadata(episode)
  episode_dict = episode_metadata_fn(episode_dict)
  if episode_dict is None:
    return None
  steps = _generate_steps(episode, step_fn)

  episode_dict['steps'] = steps
  return episode_dict


def generate_episodes(
    tag_reader: Any,
    episode_metadata_fn: Callable[
        [Dict[str, Any]], Optional[Dict[str, Any]]
    ] = lambda x: x,
    step_fn: Callable[[Dict[str, Any]], Dict[str, Any]] = lambda x: x,
) -> Generator[Dict[str, Any], None, None]:
  """Generates episodes by reading them from the tag_reader.

  Args:
    tag_reader: Envlogger reader (envlogger.Reader).
    episode_metadata_fn: function to process the episode metadata. If it returns
      None, the episode is skipped.
    step_fn: function that gets an RLDS step and returns the processed step.

  Yields:
    A dictionary representing one episode, including metadata and steps. Steps
    are represented by a nested dictionary where each nested field is a list.
  """
  for index, episode_metadata in enumerate(tag_reader.episode_metadata()):
    episode = tag_reader.episodes[index]
    episode_dict = get_episode_dict(
        episode, episode_metadata, episode_metadata_fn, step_fn
    )
    if episode_dict:
      yield episode_dict


def _build_empty_step(
    step: Any, step_fn: Callable[[Dict[str, Any]], Dict[str, Any]]
) -> Dict[str, Any]:
  """Builds a step with the same shape and dtype of an RLDS step."""
  rlds_step = tree_utils.map_structure(np.zeros_like, _get_step_metadata(step))
  rlds_step['observation'] = tree_utils.map_structure(
      np.zeros_like, step.timestep.observation
  )
  rlds_step['action'] = tree_utils.map_structure(np.zeros_like, step.action)
  rlds_step['reward'] = np.zeros_like(step.timestep.reward)
  rlds_step['discount'] = np.zeros_like(step.timestep.discount)
  rlds_step['is_terminal'] = False
  rlds_step['is_first'] = False
  rlds_step['is_last'] = False

  return step_fn(rlds_step)


def _build_step(
    prev_step: Any,
    step: Any,
    step_fn: Callable[[Dict[str, Any]], Dict[str, Any]],
) -> Dict[str, Any]:
  """Builds an RLDS step from two envlogger steps."""
  rlds_step = _get_step_metadata(prev_step)
  rlds_step['observation'] = prev_step.timestep.observation
  rlds_step['action'] = step.action
  rlds_step['reward'] = step.timestep.reward
  rlds_step['discount'] = step.timestep.discount
  rlds_step['is_terminal'] = False
  rlds_step['is_first'] = prev_step.timestep.first()
  rlds_step['is_last'] = prev_step.timestep.last()

  return step_fn(rlds_step)


def _build_last_step(
    prev_step: Any, step_fn: Callable[[Dict[str, Any]], Dict[str, Any]]
) -> Dict[str, Any]:
  """Builds the last RLDS step from an envlogger step."""
  # We append the observation of the final step (action and reward were
  # included in the previous step.
  # The terminal flag is inferred like in termination(), truncation()
  # from dm_env/_environment.py
  is_terminal = prev_step.timestep.last() and prev_step.timestep.discount == 0.0
  rlds_step = _get_step_metadata(prev_step)
  rlds_step['observation'] = prev_step.timestep.observation
  rlds_step['is_terminal'] = is_terminal
  rlds_step['is_first'] = prev_step.timestep.first()
  rlds_step['is_last'] = True
  # Discount, action and reward are meaningless in the last step
  rlds_step['action'] = tree_utils.map_structure(
      np.zeros_like, prev_step.action
  )
  rlds_step['reward'] = np.zeros_like(prev_step.timestep.reward)
  rlds_step['discount'] = np.zeros_like(prev_step.timestep.discount)

  return step_fn(rlds_step)


def _generate_steps(
    episode: Sequence[Any], step_fn: Callable[[Dict[str, Any]], Dict[str, Any]]
) -> Dict[str, Any]:
  """Constructs a dictionary of steps for the given episode.

  Args:
    episode: Sequence of steps (envlogger.StepData). Assumes that there is at
      least one step.
    step_fn: function that gets an RLDS step and returns the processed step.

  Returns:
    Nested dictionary with the steps of one episode. Steps
    are represented by a nested dictionary where each nested field is a list.
  """
  if len(episode) < 2:
    step = _build_last_step(episode[0], step_fn)
    return _to_nested_list(step)
  # We use episode[1] here because in the first steps some of the fields(like
  # the action, migt be None, so the shape is unknown).
  empty_step = _build_empty_step(episode[1], step_fn)
  steps = _empty_nested_list(empty_step)

  prev_step = None
  for step in episode:
    if prev_step is not None:
      rlds_step = _build_step(prev_step, step, step_fn)
      for k, v in rlds_step.items():
        steps[k] = _append_nested(steps[k], v)
    prev_step = step

  if prev_step is not None:
    last_step = _build_last_step(prev_step, step_fn)
    for k, v in last_step.items():
      steps[k] = _append_nested(steps[k], v)
  return steps


def _append_nested(
    collection: Any,
    data: Union[np.ndarray, List[Any], Tuple[Any, ...], Dict[str, Any]],
) -> Any:
  """Appends data to the more nested dimension of collection.

  Args:
    collection: collection constructed by appending elements like data to the
      most nested dimension. It is of type Union[np.ndarray, List[Any],
      Tuple[Any, ...], Dict[str, Any]] but using this causes type errors because
      it considers that we are trying to append to a dict or a tuple.
    data: data logged by DM Env Logger that can be a scalar, array, dict or
      tuple.

  Returns:
    Collection with data appended to the nested dimension.
  """
  if isinstance(data, dict):
    for k in collection:
      _append_nested(collection[k], data[k])
    return collection
  elif isinstance(data, tuple):
    return tuple(
        _append_nested(collection[index], value)
        for index, value in enumerate(data)
    )
  else:
    collection.append(data)
    return collection


def _apply_recursive(
    data: Union[np.ndarray, List[Any], Tuple[Any, ...], Dict[str, Any]],
    fn: Callable[[Any], Any],
) -> Union[List[Any], Tuple[Any, ...], Dict[str, Any]]:
  """Applies a function recursively to the nested dimensions of the input.

  Args:
    data: scalar, array, list, tuple or dictionary.
    fn: function to apply to the most nested dimension.

  Returns:
    item with the same shape of data but with fn applied to the most nested
    dimension
  """
  if isinstance(data, dict):
    return {k: _apply_recursive(data[k], fn) for k in data}
  elif isinstance(data, tuple):
    return tuple(_apply_recursive(x, fn) for x in data)
  else:
    return fn(data)


def _empty_nested_list(data: Dict[str, Any]) -> Dict[str, Any]:
  """Creates an element like data but with empty lists in the nested dimensions.

  Args:
    data: dictionary.

  Returns:
    item with the same shape of data but with empty lists in its nested
    dimensions.
  """

  return {k: _apply_recursive(v, lambda x: []) for k, v in data.items()}


def _to_nested_list(data: Dict[str, Any]) -> Dict[str, Any]:
  """Transforms data to lists of length 1 in the nested dimensions.

  Args:
    data: dictionary.

  Returns:
    item with the same shape of data but with lists of length 1 in its nested
    dimensions.
  """
  return {k: _apply_recursive(v, lambda x: [x]) for k, v in data.items()}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for collection_markdown_builder."""
import textwrap

import pytest

import tensorflow_datasets as tfds
from tensorflow_datasets import testing
from tensorflow_datasets.scripts.documentation import collection_markdown_builder


@pytest.fixture(scope='session')
def dummy_dc_loader() -> tfds.core.DatasetCollectionLoader:
  return tfds.core.DatasetCollectionLoader(
      collection=testing.DummyDatasetCollection()
  )


def test_collection_homepage_section(
    dummy_dc_loader: tfds.core.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  homepage_section = collection_markdown_builder.CollectionHomepageSection()
  assert not homepage_section.content(loader=dummy_dc_loader)


def test_collection_description_section(
    dummy_dc_loader: tfds.core.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  description_section = (
      collection_markdown_builder.CollectionDescriptionSection()
  )
  expected_description = dummy_dc_loader.collection.info.description
  assert (
      description_section.content(loader=dummy_dc_loader)
      == expected_description
  )


def test_collection_citation_section(
    dummy_dc_loader: tfds.core.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  citation_section = collection_markdown_builder.CollectionCitationSection()
  expected_citation = textwrap.dedent(
      f"""
            ```
            {tfds.core.utils.indent(dummy_dc_loader.collection.info.citation, '            ')}
            ```
            """
  )
  assert citation_section.content(loader=dummy_dc_loader) == expected_citation


def test_collection_versions_section(
    dummy_dc_loader: tfds.core.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  versions_section = collection_markdown_builder.CollectionVersionSection()
  expected_lines = [
      '    * `1.0.0`: notes 1.0.0',
      '    * `1.1.0`: notes 1.1.0',
      '    * **`2.0.0`** (default): notes 2.0.0',
  ]
  assert versions_section.content(loader=dummy_dc_loader) == '\n'.join(
      expected_lines
  )


def test_collection_datasets_section(
    dummy_dc_loader: tfds.core.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  datasets_section = collection_markdown_builder.CollectionDatasetsSection()
  expected_lines = [
      (
          '    * `a`:'
          ' [`a/c:1.3.5`](https://www.tensorflow.org/datasets/catalog/a#c)'
      ),
      (
          '    * `b`:'
          ' [`b/d:2.4.8`](https://www.tensorflow.org/datasets/catalog/b#d)'
      ),
      (
          '    * `c`:'
          ' [`c/e:3.5.7`](https://www.tensorflow.org/datasets/catalog/c#e)'
      ),
  ]
  assert datasets_section.content(loader=dummy_dc_loader) == '\n'.join(
      expected_lines
  )


def test_get_collection_markdown_string(
    dummy_dc_loader: tfds.core.DatasetCollectionLoader,
):  # pylint: disable=redefined-outer-name
  doc = collection_markdown_builder.get_collection_markdown_string(
      collection=dummy_dc_loader
  )

  assert 'dummy_dataset_collection' in doc  # Collection heading.
  assert 'my description' in doc  # Collection description.
  assert '@misc{citekey' in doc  # Collection citation.
  assert (
      '[`c/e:3.5.7`](https://www.tensorflow.org/datasets/catalog/c#e)' in doc
  )  # Collection dataset.



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.utils.image_utils."""

import os

import numpy as np
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core.utils import image_utils

# pylint: disable=bad-whitespace
SIX_PIXELS = [
    [[0, 255, 0], [255, 0, 0], [255, 0, 255]],
    [[0, 0, 255], [255, 255, 0], [126, 127, 128]],
]

SIX_PIXELS_JPEG = [
    [[158, 161, 92], [76, 79, 10], [180, 57, 181]],
    [[33, 36, 0], [229, 232, 163], [201, 78, 202]],
]
# pylint: enable=bad-whitespace


class ImageUtilsTest(testing.TestCase):

  def _get_image(self, name):
    path = os.path.join(self.test_data, name)
    with tf.io.gfile.GFile(path, 'rb') as img_f:
      return img_f.read()

  def test_decode_image(self):
    image = self._get_image('6pixels.png')
    np_image = image_utils.decode_image(image)
    np.testing.assert_array_equal(np_image, SIX_PIXELS)

  def test_png_to_jpeg(self):
    image = self._get_image('6pixels.png')
    jpeg = image_utils.png_to_jpeg(image)
    image_np_jpeg = image_utils.decode_image(jpeg)
    np.testing.assert_array_equal(image_np_jpeg, SIX_PIXELS_JPEG)

  def test_png_4chan_to_jpeg(self):
    image = self._get_image('6pixels_4chan.png')
    jpeg = image_utils.png_to_jpeg(image)
    self.assertIsInstance(jpeg, bytes)
    image_np_jpeg = image_utils.decode_image(jpeg)
    np.testing.assert_array_equal(image_np_jpeg, SIX_PIXELS_JPEG)

  def test_jpeg_cmyk_to_rgb(self):
    image = self._get_image('6pixels_cmyk.jpeg')
    new_image = image_utils.jpeg_cmyk_to_rgb(image, quality=100)
    self.assertIsInstance(new_image, bytes)
    self.assertNotEqual(image, new_image)
    # Converting between color systems is not bijective, so high rtol.
    original_np_image = image_utils.decode_image(image)
    new_np_image = image_utils.decode_image(new_image)
    np.testing.assert_allclose(original_np_image, new_np_image, rtol=10)

  def test_apply_colormap(self):
    assert image_utils.get_colormap().shape == (256, 3)

    gray_img = np.array(
        [
            [[0], [2]],
            [[1234], [2]],
        ],
        dtype=np.uint16,
    )
    assert gray_img.shape == (2, 2, 1)
    colored_img = image_utils.apply_colormap(gray_img)
    assert colored_img.shape == (2, 2, 3)
    assert colored_img.dtype == np.uint8

    np.testing.assert_array_equal(
        colored_img,
        [
            [[0, 0, 0], [55, 126, 184]],
            [[71, 84, 183], [55, 126, 184]],
        ],
    )


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wikipedia-based Image Text (WIT) Dataset for the Kaggle competition."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.vision_language.wit_kaggle import wit_kaggle


class WitKaggleTestTrain(tfds.testing.DatasetBuilderTestCase):
  """Tests for wit_kaggle dataset train split."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['train_with_extended_features']
  DATASET_CLASS = wit_kaggle.WitKaggle
  SKIP_CHECKSUMS = True  # All data is manually downloaded.

  SPLITS = {'train_with_extended_features': 2}
  DL_EXTRACT_ONLY_RESULT = {
      'samples': ['wit_kaggle/train/ZIP.unzipped_train'],
      'images': 'wit_kaggle/train/image_data_train',
  }


class WitKaggleTestTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for wit_kaggle dataset test split."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['test_without_gold']
  DATASET_CLASS = wit_kaggle.WitKaggle
  SKIP_CHECKSUMS = True  # All data is manually downloaded.

  SPLITS = {'test_without_gold': 1}
  DL_EXTRACT_ONLY_RESULT = {
      'samples': ['wit_kaggle/test/ZIP.unzipped_test'],
      'images': 'wit_kaggle/test/image_data_test',
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""iNaturalist2021 dataset."""

import json
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
The iNaturalist dataset 2021 contains a total of 10,000 species. 
The full training dataset contains nearly 2.7M images. 
To make the dataset more accessible we have also created a "mini" training 
dataset with 50 examples per species for a total of 500K images. The full 
training `train` split overlaps with the `mini` split. The val set contains for
each species 10 validation images (100K in total). There are a total of 500,000 
test images in the `public_test` split (without ground-truth labels).
"""

_CITATION = r"""\
@misc{inaturalist21,
    Howpublished = {~\url{https://github.com/visipedia/inat_comp/tree/master/2021}},
    Title = {{iNaturalist} 2021 competition dataset.},
    Year = {2021},
    key = {{iNaturalist} 2021 competition dataset},
    }
"""

_URL = 'https://ml-inat-competition-datasets.s3.amazonaws.com/2021'
_HOMEPAGE = 'https://github.com/visipedia/inat_comp/tree/master/2021'
_SPLIT_FILENAMES = {
    'train': 'train',
    'mini': 'train_mini',
    'val': 'val',
    'test': 'public_test',
}


class INaturalist2021(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for iNaturalist 2021 Competition dataset."""

  VERSION = tfds.core.Version('2.0.1')
  RELEASE_NOTES = {
      '1.0.0': 'Initial release.',
      '2.0.0': 'Update: Class indices follow the order in the JSON train file.',
      '2.0.1': 'Update: Include the example id as provided in the JSON files.',
  }

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'id': tfds.features.Scalar(dtype=np.int64),
            'file_id': tfds.features.Text(),
            'image': tfds.features.Image(shape=(None, None, 3)),
            'label': tfds.features.ClassLabel(
                names_file=tfds.core.tfds_path(
                    os.path.join(
                        'image_classification',
                        'i_naturalist2021',
                        'i_naturalist2021_labels_ordered_by_json_id.txt',
                    )
                )
            ),
            'supercategory': tfds.features.ClassLabel(
                names_file=tfds.core.tfds_path(
                    os.path.join(
                        'image_classification',
                        'i_naturalist2021',
                        'i_naturalist2021_supercategories.txt',
                    )
                )
            ),
        }),
        supervised_keys=('image', 'label'),
        homepage=_HOMEPAGE,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""

    split_downloads = {}
    for split, split_file in _SPLIT_FILENAMES.items():
      split_downloads[f'{split}_img'] = tfds.download.Resource(
          url=f'{_URL}/{split_file}.tar.gz',
          extract_method=tfds.download.ExtractMethod.NO_EXTRACT,
      )
      split_downloads[f'{split}_json'] = f'{_URL}/{split_file}.json.tar.gz'

    output_paths = dl_manager.download_and_extract(split_downloads)
    generate_dict = {}
    for split, split_file in _SPLIT_FILENAMES.items():
      generate_dict[split] = self._generate_examples(
          images_archive=dl_manager.iter_archive(output_paths[f'{split}_img']),
          json_file=os.path.join(
              output_paths[f'{split}_json'], f'{split_file}.json'
          ),
      )

    return generate_dict

  def _generate_examples(self, images_archive, json_file):
    """Generate examples."""
    # Training and validation images.
    with epath.Path(json_file).open('r') as f:
      inat_json = json.load(f)

    def _format(label: str):
      return label.lower().replace(' ', '_')

    def _get_annotation(idx, image_id):
      if 'annotations' in inat_json:
        annotation = inat_json['annotations'][idx]
        assert annotation['image_id'] == image_id
        cat = inat_json['categories'][annotation['category_id']]
        category = _format(cat['name'])
        supercategory = _format(cat['supercategory'])
      else:
        category, supercategory = -1, -1
      return category, supercategory

    fileid2data = {}
    for idx, image in enumerate(inat_json['images']):
      category, supercategory = _get_annotation(idx, image['id'])
      fileid = os.path.basename(image['file_name']).split('.')[0]
      fileid2data[fileid] = {
          'id': image['id'],
          'file_id': fileid,
          'label': category,
          'supercategory': supercategory,
      }

    # Read tar.gz file containing images and yield relevant examples.
    for fpath, fobj in images_archive:
      fileid = os.path.basename(fpath).split('.')[0]
      if fileid in fileid2data:
        data = fileid2data[fileid].copy()
        data['image'] = fobj
        yield data['id'], data



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: skip-file

# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: dataset_info.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tensorflow_datasets.core.proto import feature_generated_pb2 as feature__pb2
from tensorflow_metadata.proto.v0 import schema_pb2 as tensorflow__metadata_dot_proto_dot_v0_dot_schema__pb2
from tensorflow_metadata.proto.v0 import statistics_pb2 as tensorflow__metadata_dot_proto_dot_v0_dot_statistics__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x12\x64\x61taset_info.proto\x12\x13tensorflow_datasets\x1a\rfeature.proto\x1a)tensorflow_metadata/proto/v0/schema.proto\x1a-tensorflow_metadata/proto/v0/statistics.proto"\x1f\n\x0f\x44\x61tasetLocation\x12\x0c\n\x04urls\x18\x01'
    b' \x03(\t"\xb8\x01\n\tSplitInfo\x12\x0c\n\x04name\x18\x01'
    b' \x01(\t\x12\x12\n\nnum_shards\x18\x02'
    b' \x01(\x03\x12\x15\n\rshard_lengths\x18\x04'
    b' \x03(\x03\x12\x11\n\tnum_bytes\x18\x05'
    b' \x01(\x03\x12\x44\n\nstatistics\x18\x03'
    b' \x01(\x0b\x32\x30.tensorflow.metadata.v0.DatasetFeatureStatistics\x12\x19\n\x11\x66ilepath_template\x18\x06'
    b' \x01(\t"\xf3\x03\n\x0eSupervisedKeys\x12\x11\n\x05input\x18\x01'
    b' \x01(\tB\x02\x18\x01\x12\x12\n\x06output\x18\x02'
    b' \x01(\tB\x02\x18\x01\x12\x38\n\x05tuple\x18\x03'
    b' \x01(\x0b\x32).tensorflow_datasets.SupervisedKeys.Tuple\x1a@\n\x05Tuple\x12\x37\n\x05items\x18\x01'
    b' \x03(\x0b\x32(.tensorflow_datasets.SupervisedKeys.Nest\x1a\x9f\x01\n\x04\x44ict\x12@\n\x04\x64ict\x18\x01'
    b' \x03(\x0b\x32\x32.tensorflow_datasets.SupervisedKeys.Dict.DictEntry\x1aU\n\tDictEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12\x37\n\x05value\x18\x02'
    b' \x01(\x0b\x32(.tensorflow_datasets.SupervisedKeys.Nest:\x02\x38\x01\x1a\x9b\x01\n\x04Nest\x12\x15\n\x0b\x66\x65\x61ture_key\x18\x01'
    b' \x01(\tH\x00\x12:\n\x05tuple\x18\x02'
    b' \x01(\x0b\x32).tensorflow_datasets.SupervisedKeys.TupleH\x00\x12\x38\n\x04\x64ict\x18\x03'
    b' \x01(\x0b\x32(.tensorflow_datasets.SupervisedKeys.DictH\x00\x42\x06\n\x04nest"%\n\x12RedistributionInfo\x12\x0f\n\x07license\x18\x01'
    b' \x01(\t"\x8f\x02\n\x10\x44\x61taSourceAccess\x12\x1b\n\x13\x61\x63\x63\x65ss_timestamp_ms\x18\x01'
    b' \x01(\x03\x12\x36\n\x0b\x66ile_system\x18\x02'
    b' \x01(\x0b\x32\x1f.tensorflow_datasets.FileSystemH\x00\x12\x32\n\tsql_query\x18\x03'
    b' \x01(\x0b\x32\x1d.tensorflow_datasets.SqlQueryH\x00\x12\x41\n\x0ctfds_dataset\x18\x04'
    b' \x01(\x0b\x32).tensorflow_datasets.TfdsDatasetReferenceH\x00\x12%\n\x03url\x18\x05'
    b' \x01(\x0b\x32\x18.tensorflow_datasets.UrlB\x08\n\x06source"\x1a\n\nFileSystem\x12\x0c\n\x04path\x18\x01'
    b' \x01(\t"$\n\x03Url\x12\x0b\n\x03url\x18\x01'
    b' \x01(\t\x12\x10\n\x08\x63hecksum\x18\x02'
    b' \x01(\t"\x1d\n\x08SqlQuery\x12\x11\n\tsql_query\x18\x01'
    b' \x01(\t"m\n\x14TfdsDatasetReference\x12\x0c\n\x04name\x18\x01'
    b' \x01(\t\x12\x0e\n\x06\x63onfig\x18\x02'
    b' \x01(\t\x12\x0f\n\x07version\x18\x03'
    b' \x01(\t\x12\x10\n\x08\x64\x61ta_dir\x18\x04'
    b' \x01(\t\x12\x14\n\x0c\x64s_namespace\x18\x05'
    b' \x01(\t"\xb4\x07\n\x0b\x44\x61tasetInfo\x12\x0c\n\x04name\x18\x01'
    b' \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02'
    b' \x01(\t\x12\x0f\n\x07version\x18\t \x01(\t\x12I\n\rrelease_notes\x18\x12'
    b' \x03(\x0b\x32\x32.tensorflow_datasets.DatasetInfo.ReleaseNotesEntry\x12\x13\n\x0b\x63onfig_name\x18\r'
    b' \x01(\t\x12\x1a\n\x12\x63onfig_description\x18\x0e'
    b' \x01(\t\x12\x13\n\x0b\x63onfig_tags\x18\x15'
    b' \x03(\t\x12.\n\x08\x66\x65\x61tures\x18\x13'
    b' \x01(\x0b\x32\x1c.tensorflow_datasets.Feature\x12\x10\n\x08\x63itation\x18\x03'
    b' \x01(\t\x12\x19\n\rsize_in_bytes\x18\x04'
    b' \x01(\x03\x42\x02\x18\x01\x12\x15\n\rdownload_size\x18\x0c'
    b' \x01(\x03\x12\x36\n\x08location\x18\x05'
    b' \x01(\x0b\x32$.tensorflow_datasets.DatasetLocation\x12W\n\x12\x64ownload_checksums\x18\n'
    b' \x03(\x0b\x32\x37.tensorflow_datasets.DatasetInfo.DownloadChecksumsEntryB\x02\x18\x01\x12.\n\x06schema\x18\x06'
    b' \x01(\x0b\x32\x1e.tensorflow.metadata.v0.Schema\x12.\n\x06splits\x18\x07'
    b' \x03(\x0b\x32\x1e.tensorflow_datasets.SplitInfo\x12<\n\x0fsupervised_keys\x18\x08'
    b' \x01(\x0b\x32#.tensorflow_datasets.SupervisedKeys\x12\x44\n\x13redistribution_info\x18\x0b'
    b" \x01(\x0b\x32'.tensorflow_datasets.RedistributionInfo\x12\x13\n\x0bmodule_name\x18\x0f"
    b' \x01(\t\x12\x19\n\x11\x64isable_shuffling\x18\x10'
    b' \x01(\x08\x12\x13\n\x0b\x66ile_format\x18\x11'
    b' \x01(\t\x12\x43\n\x14\x64\x61ta_source_accesses\x18\x14'
    b' \x03(\x0b\x32%.tensorflow_datasets.DataSourceAccess\x1a\x33\n\x11ReleaseNotesEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12\r\n\x05value\x18\x02'
    b' \x01(\t:\x02\x38\x01\x1a\x38\n\x16\x44ownloadChecksumsEntry\x12\x0b\n\x03key\x18\x01'
    b' \x01(\t\x12\r\n\x05value\x18\x02'
    b' \x01(\t:\x02\x38\x01\x42\x03\xf8\x01\x01\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, 'dataset_info_pb2', globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\370\001\001'
  _SUPERVISEDKEYS_DICT_DICTENTRY._options = None
  _SUPERVISEDKEYS_DICT_DICTENTRY._serialized_options = b'8\001'
  _SUPERVISEDKEYS.fields_by_name['input']._options = None
  _SUPERVISEDKEYS.fields_by_name['input']._serialized_options = b'\030\001'
  _SUPERVISEDKEYS.fields_by_name['output']._options = None
  _SUPERVISEDKEYS.fields_by_name['output']._serialized_options = b'\030\001'
  _DATASETINFO_RELEASENOTESENTRY._options = None
  _DATASETINFO_RELEASENOTESENTRY._serialized_options = b'8\001'
  _DATASETINFO_DOWNLOADCHECKSUMSENTRY._options = None
  _DATASETINFO_DOWNLOADCHECKSUMSENTRY._serialized_options = b'8\001'
  _DATASETINFO.fields_by_name['size_in_bytes']._options = None
  _DATASETINFO.fields_by_name['size_in_bytes']._serialized_options = b'\030\001'
  _DATASETINFO.fields_by_name['download_checksums']._options = None
  _DATASETINFO.fields_by_name['download_checksums']._serialized_options = (
      b'\030\001'
  )
  _DATASETLOCATION._serialized_start = 148
  _DATASETLOCATION._serialized_end = 179
  _SPLITINFO._serialized_start = 182
  _SPLITINFO._serialized_end = 366
  _SUPERVISEDKEYS._serialized_start = 369
  _SUPERVISEDKEYS._serialized_end = 868
  _SUPERVISEDKEYS_TUPLE._serialized_start = 484
  _SUPERVISEDKEYS_TUPLE._serialized_end = 548
  _SUPERVISEDKEYS_DICT._serialized_start = 551
  _SUPERVISEDKEYS_DICT._serialized_end = 710
  _SUPERVISEDKEYS_DICT_DICTENTRY._serialized_start = 625
  _SUPERVISEDKEYS_DICT_DICTENTRY._serialized_end = 710
  _SUPERVISEDKEYS_NEST._serialized_start = 713
  _SUPERVISEDKEYS_NEST._serialized_end = 868
  _REDISTRIBUTIONINFO._serialized_start = 870
  _REDISTRIBUTIONINFO._serialized_end = 907
  _DATASOURCEACCESS._serialized_start = 910
  _DATASOURCEACCESS._serialized_end = 1181
  _FILESYSTEM._serialized_start = 1183
  _FILESYSTEM._serialized_end = 1209
  _URL._serialized_start = 1211
  _URL._serialized_end = 1247
  _SQLQUERY._serialized_start = 1249
  _SQLQUERY._serialized_end = 1278
  _TFDSDATASETREFERENCE._serialized_start = 1280
  _TFDSDATASETREFERENCE._serialized_end = 1389
  _DATASETINFO._serialized_start = 1392
  _DATASETINFO._serialized_end = 2340
  _DATASETINFO_RELEASENOTESENTRY._serialized_start = 2231
  _DATASETINFO_RELEASENOTESENTRY._serialized_end = 2282
  _DATASETINFO_DOWNLOADCHECKSUMSENTRY._serialized_start = 2284
  _DATASETINFO_DOWNLOADCHECKSUMSENTRY._serialized_end = 2340
# @@protoc_insertion_point(module_scope)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for extractor."""

import os
from unittest import mock

import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core.download import extractor
from tensorflow_datasets.core.download import resource as resource_lib

BZIP2 = resource_lib.ExtractMethod.BZIP2
GZIP = resource_lib.ExtractMethod.GZIP
NO_EXTRACT = resource_lib.ExtractMethod.NO_EXTRACT
TAR = resource_lib.ExtractMethod.TAR
TAR_GZ = resource_lib.ExtractMethod.TAR_GZ
ZIP = resource_lib.ExtractMethod.ZIP
TAR_STREAM = resource_lib.ExtractMethod.TAR_STREAM
TAR_GZ_STREAM = resource_lib.ExtractMethod.TAR_GZ_STREAM


def _read(path):
  with tf.io.gfile.GFile(path, 'rb') as f:
    return f.read()


class ExtractorTest(testing.TestCase):

  @classmethod
  def setUpClass(cls):
    super(ExtractorTest, cls).setUpClass()
    cls.f1_content = _read(os.path.join(cls.test_data, '6pixels.png'))
    cls.f2_content = _read(os.path.join(cls.test_data, 'foo.csv'))

  def setUp(self):
    super(ExtractorTest, self).setUp()
    self.extractor = extractor._Extractor()
    self.extractor._pbar_path = mock.MagicMock()
    # Where archive will be extracted:
    self.to_path = os.path.join(self.tmp_dir, 'extracted_arch')
    # Obviously it must not exist before test runs:
    self.assertFalse(tf.io.gfile.exists(self.to_path))

    self.result_path = os.path.join(self.to_path, '6pixels.png')

  def test_unknown_method(self):
    with self.assertRaises(ValueError):
      self.extractor.extract('from/path', NO_EXTRACT, 'to/path')

  def _test_extract(self, method, archive_name, expected_files):
    from_path = os.path.join(self.test_data, 'archives', archive_name)
    path = self.extractor.extract(from_path, method, self.to_path).get()
    self.assertIsInstance(path, os.PathLike)
    for name, content in expected_files.items():
      path = os.path.join(self.to_path, name)
      self.assertEqual(_read(path), content, 'File %s has bad content.' % path)

  def test_zip(self):
    self._test_extract(
        ZIP,
        'arch1.zip',
        {'6pixels.png': self.f1_content, 'foo.csv': self.f2_content},
    )

  def test_tar(self):
    self._test_extract(
        TAR,
        'arch1.tar',
        {'6pixels.png': self.f1_content, 'foo.csv': self.f2_content},
    )

  def test_targz(self):
    self._test_extract(
        TAR_GZ,
        'arch1.tar.gz',
        {'6pixels.png': self.f1_content, 'foo.csv': self.f2_content},
    )

  def test_tar_stream(self):
    self._test_extract(
        TAR_STREAM,
        'arch1.tar',
        {'6pixels.png': self.f1_content, 'foo.csv': self.f2_content},
    )

  def test_targz_stream(self):
    self._test_extract(
        TAR_GZ_STREAM,
        'arch1.tar.gz',
        {'6pixels.png': self.f1_content, 'foo.csv': self.f2_content},
    )

  def test_gzip(self):
    from_path = os.path.join(self.test_data, 'archives', 'arch1.tar.gz')
    self.extractor.extract(from_path, GZIP, self.to_path).get()
    arch1_path = os.path.join(self.test_data, 'archives', 'arch1.tar')
    self.assertEqual(_read(self.to_path), _read(arch1_path))

  def test_gzip2(self):
    # Same as previous test, except it is not a .tar.gz, but a .gz.
    from_path = os.path.join(self.test_data, 'archives', 'foo.csv.gz')
    self.extractor.extract(from_path, GZIP, self.to_path).get()
    foo_csv_path = os.path.join(self.test_data, 'foo.csv')
    self.assertEqual(_read(self.to_path), _read(foo_csv_path))

  def test_bzip2(self):
    from_path = os.path.join(self.test_data, 'archives', 'foo.csv.bz2')
    self.extractor.extract(from_path, BZIP2, self.to_path).get()
    foo_csv_path = os.path.join(self.test_data, 'foo.csv')
    self.assertEqual(_read(self.to_path), _read(foo_csv_path))

  def test_absolute_path(self):
    # There is a file with absolute path (ignored) + a file named "foo".
    self._test_extract(TAR, 'absolute_path.tar', {'foo': b'bar\n'})

  def test_wrong_method(self):
    from_path = os.path.join(self.test_data, 'archives', 'foo.csv.gz')
    promise = self.extractor.extract(from_path, ZIP, self.to_path)
    expected_msg = 'File is not a zip file'
    with self.assertRaisesWithPredicateMatch(
        extractor.ExtractError, expected_msg
    ):
      promise.get()


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualizer utils."""

from tensorflow_datasets.core.visualization.graph_visualizer import GraphVisualizer
from tensorflow_datasets.core.visualization.graph_visualizer import GraphVisualizerMetadataDict
from tensorflow_datasets.core.visualization.image_visualizer import ImageGridVisualizer
from tensorflow_datasets.core.visualization.show_examples import show_examples
from tensorflow_datasets.core.visualization.show_examples import show_statistics
from tensorflow_datasets.core.visualization.visualizer import Visualizer

__all__ = [
    "ImageGridVisualizer",
    "GraphVisualizer",
    "GraphVisualizerMetadataDict",
    "show_examples",
    "show_statistics",
    "Visualizer",
]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Methods to retrieve and store size/checksums associated to URLs."""

import dataclasses
import hashlib
import io
from typing import Any, Dict, Iterable, Optional

from absl import logging
from etils import epath
from tensorflow_datasets.core import utils

_CHECKSUM_DIRS = [
    utils.tfds_path() / 'url_checksums',
]
_CHECKSUM_SUFFIX = '.txt'


@dataclasses.dataclass(eq=True)
class UrlInfo:
  """Small wrapper around the url metadata (checksum, size).

  Attributes:
    size: Download size of the file
    checksum: Checksum of the file
    filename: Name of the file
  """

  size: utils.Size
  checksum: str
  # We exclude the filename from `__eq__` for backward compatibility
  # Two checksums are equals even if filename is unknown or different.
  filename: Optional[str] = dataclasses.field(compare=False)

  def asdict(self) -> Dict[str, Any]:
    """Returns the dict representation of the dataclass."""
    return dataclasses.asdict(self)


def compute_url_info(
    path: epath.PathLike,
    checksum_cls=hashlib.sha256,
) -> UrlInfo:
  """Locally compute size, checksums of the given file."""
  path = epath.Path(path)

  checksum = checksum_cls()
  size = 0
  with path.open('rb') as f:
    while True:
      block = f.read(io.DEFAULT_BUFFER_SIZE)
      size += len(block)
      if not block:
        break
      checksum.update(block)

  return UrlInfo(
      checksum=checksum.hexdigest(),  # base64 digest would have been better.
      size=utils.Size(size),
      filename=path.name,
  )


def add_checksums_dir(checksums_dir: str) -> None:
  """Registers a new checksums dir.

  This function allow external datasets not present in the tfds repository to
  define their own checksums_dir containing the dataset downloads checksums.

  Note: When redistributing your dataset, you should distribute the checksums
  files with it and set `add_checksums_dir` when the user is importing your
  `my_dataset.py`.

  ```
  # Set-up the folder containing the 'my_dataset.txt' checksums.
  checksum_dir = os.path.join(os.path.dirname(__file__), 'checksums/')
  checksum_dir = os.path.normpath(checksum_dir)

  # Add the checksum dir (will be executed when the user import your dataset)
  tfds.download.add_checksums_dir(checksum_dir)

  class MyDataset(tfds.core.DatasetBuilder):
    ...
  ```

  Args:
    checksums_dir: `str`, checksums dir to add to the registry
  """
  logging.warning(
      '`tfds.core.add_checksums_dir` is deprecated. Refactor dataset in '
      'self-contained folders (`my_dataset/` folder containing '
      'my_dataset.py, my_dataset_test.py, dummy_data/, checksums.tsv). '
      'The checksum file will be automatically detected. More info at: '
      'https://www.tensorflow.org/datasets/add_dataset'
  )
  if checksums_dir in _CHECKSUM_DIRS:  # Avoid duplicate
    return
  _CHECKSUM_DIRS.append(checksums_dir)


@utils.memoize()
def _checksum_paths() -> Dict[str, epath.Path]:
  """Returns dict {'dataset_name': 'path/to/checksums/file'}."""
  dataset2path = {}
  for dir_path in _CHECKSUM_DIRS:
    if isinstance(dir_path, str):
      dir_path = epath.Path(dir_path)
    if not dir_path.exists():
      pass
    for file_path in dir_path.iterdir():
      if not file_path.name.endswith(_CHECKSUM_SUFFIX):
        continue
      dataset_name = file_path.name[: -len(_CHECKSUM_SUFFIX)]
      dataset2path[dataset_name] = file_path
  return dataset2path


def _parse_url_infos(checksums_file: Iterable[str]) -> Dict[str, UrlInfo]:
  """Returns {URL: (size, checksum)}s stored within given file."""
  url_infos = {}
  for line in checksums_file:
    line = line.strip()  # Remove the trailing '\r' on Windows OS.
    if not line or line.startswith('#'):
      continue
    values = line.split('\t')
    if len(values) == 1:  # not enough values to unpack (legacy files)
      # URL might have spaces inside, but size and checksum will not.
      values = line.rsplit(' ', 2)
    if len(values) == 4:
      url, size, checksum, filename = values
    elif len(values) == 3:
      url, size, checksum = values
      filename = None
    else:
      raise AssertionError(f'Error parsing checksums: {values}')
    url_infos[url] = UrlInfo(
        size=utils.Size(size),
        checksum=checksum,
        filename=filename,
    )
  return url_infos


@utils.memoize()
def get_all_url_infos() -> Dict[str, UrlInfo]:
  """Returns dict associating URL to UrlInfo."""
  url_infos = {}
  for path in _checksum_paths().values():
    dataset_url_infos = load_url_infos(path)
    for url, url_info in dataset_url_infos.items():
      if url_infos.get(url, url_info) != url_info:
        raise AssertionError(
            'URL {} is registered with 2+ distinct size/checksum tuples. '
            '{} vs {}'.format(url, url_info, url_infos[url])
        )
    url_infos.update(dataset_url_infos)
  return url_infos


def load_url_infos(path: epath.PathLike) -> Dict[str, UrlInfo]:
  """Loads the checksums."""
  return _parse_url_infos(epath.Path(path).read_text().splitlines())


def save_url_infos(
    path: epath.Path,
    url_infos: Dict[str, UrlInfo],
) -> None:
  """Store given checksums and sizes for specific dataset.

  Content of file is never disgarded, only updated. This is to ensure that if
  process is killed right after first download finishes, checksums registered
  during previous runs aren't lost.

  It is the responsibility of the caller not to call function multiple times in
  parallel for a given dataset.

  Only original file content is updated. This means the entire set of new sizes
  and checksums must be given at every call.

  Args:
    path: Path to the resources.
    url_infos: dict, {url: (size_in_bytes, checksum)}.
  """
  original_data = load_url_infos(path) if path.exists() else {}
  new_data = original_data.copy()
  new_data.update(url_infos)
  # Compare filenames separately, as filename field is eq=False
  if original_data == new_data and _filenames_equal(original_data, new_data):
    return
  lines = [
      f'{url}\t{int(url_info.size)}\t{url_info.checksum}\t'
      f'{url_info.filename or ""}\n'
      for url, url_info in sorted(new_data.items())
  ]
  path.write_text(''.join(lines), encoding='UTF-8')


def _filenames_equal(
    left: Dict[str, UrlInfo],
    right: Dict[str, UrlInfo],
) -> bool:
  """Compare filenames."""
  return all(
      l.filename == r.filename for _, (l, r) in utils.zip_dict(left, right)
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility for handling flags."""

import re
from typing import List


def normalize_flags(argv: List[str]) -> List[str]:
  """Returns normalized explicit bolean flags for `absl.flags` compatibility.

  Note: Boolean flags in `absl.flags` can be specified with --bool, --nobool,
  as well as --bool=true/false (though not recommended); in `argparse_flags`,
  it only allows --bool, --nobool.

  Args:
    argv: Arguments for `main()`.
  """
  bolean_flag_patern = re.compile(r'--[\w_]+=(true|false)')

  def _normalize_flag(arg: str) -> str:
    if not bolean_flag_patern.match(arg):
      return arg
    if arg.endswith('=true'):
      return arg[: -len('=true')]  # `--flag=true` -> `--flag`
    elif arg.endswith('=false'):
      # `--flag=false` -> `--noflag`
      return '--no' + arg[len('--') : -len('=false')]
    else:
      raise AssertionError(f'Unrecognized arg: {arg}')

  return [_normalize_flag(a) for a in argv]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""gov_report dataset."""

import itertools
import json
import os
from typing import Any, Dict, Iterable, Tuple

from etils import epath
import tensorflow_datasets.public_api as tfds

_DESCRIPTION = """
Government report dataset consists of reports written by government research
 agencies including Congressional Research Service and U.S. Government
 Accountability Office.
"""

_CITATION = """
@inproceedings{
anonymous2022efficiently,
title={Efficiently Modeling Long Sequences with Structured State Spaces},
author={Anonymous},
booktitle={Submitted to The Tenth International Conference on Learning Representations },
year={2022},
url={https://openreview.net/forum?id=uYLFoz1vlAC},
note={under review}
}
"""

_HOMEPAGE = "https://gov-report-data.github.io/"
_URL = "https://drive.google.com/uc?export=download&id=1ik8uUVeIU-ky63vlnvxtfN2ZN-TUeov2"

_SUBSET_CRS = "crs"
_SUBSET_GAO = "gao"

_ID_KEY = "id"
_COMMON_KEYS = ("title", "released_date")
_CRS_KEYS = ("reports", "summary")
_GAO_KEYS = ("report", "highlight")
_GAO_ADDITIONAL_KEYS = ("url", "fastfact", "published_date")

_STYLE_DESCRIPTIONS = {
    "whitespace": (
        "Structures flattened and joined by whitespace. "
        "This is the format used by original paper"
    ),
    "html": (
        "Structures flattened and joined by newline while add html tags. "
        "Tags are only added for secition_title in a format like "
        "`<h2>xxx<h2>`."
    ),
    "json": "Structures represented as raw json.",
}


class GovReportConfig(tfds.core.BuilderConfig):
  """BuilderConfig for GovReportConfig."""

  def __init__(
      self,
      subset: str = "",
      style: str = "",
      supervised_keys: Tuple[str, str] = ("", ""),
      other_keys: Iterable[str] = (),
      description: str = "",
      **kwargs,
  ):
    super().__init__(
        name=f"{subset}_{style}",
        description=f"{description}\n{_STYLE_DESCRIPTIONS[style]}",
        **kwargs,
    )
    self.subset = subset

    self.supervised_keys = supervised_keys
    self.other_keys = tuple(other_keys) + (_ID_KEY,)
    self.all_keys = supervised_keys + self.other_keys

    self.style = style
    self.separator = " " if style == "whitepsace" else "\n"


def _configs_for_style(style: str):
  return [
      GovReportConfig(
          subset=_SUBSET_CRS,
          style=style,
          description="CRS report with summary.",
          supervised_keys=_CRS_KEYS,
          other_keys=_COMMON_KEYS,
      ),
      GovReportConfig(
          subset=_SUBSET_GAO,
          style=style,
          description="GAO report with highlight",
          supervised_keys=_GAO_KEYS,
          other_keys=_COMMON_KEYS + _GAO_ADDITIONAL_KEYS,
      ),
  ]


class GovReport(tfds.core.GeneratorBasedBuilder):
  """DatasetBuilder for gov_report dataset."""

  VERSION = tfds.core.Version("1.0.0")
  RELEASE_NOTES = {
      "1.0.0": "Initial release.",
  }
  BUILDER_CONFIGS = list(
      itertools.chain.from_iterable(
          [_configs_for_style(style) for style in _STYLE_DESCRIPTIONS]
      )
  )

  def _info(self) -> tfds.core.DatasetInfo:
    """Returns the dataset metadata."""
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict(
            {k: tfds.features.Text() for k in self.builder_config.all_keys}
        ),
        supervised_keys=self.builder_config.supervised_keys,
        homepage=_HOMEPAGE,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager: tfds.download.DownloadManager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(_URL)
    split_map = {
        tfds.Split.TRAIN: "train",
        tfds.Split.VALIDATION: "valid",
        tfds.Split.TEST: "test",
    }
    path = os.path.join(path, "gov-report")
    return {k: self._generate_examples(path, v) for k, v in split_map.items()}

  def _generate_examples(self, path: str, split: str):
    """Yields examples."""
    subset = self.builder_config.subset
    style = self.builder_config.style
    separator = self.builder_config.separator
    report_key, summary_key = self.builder_config.supervised_keys
    split_filename = os.path.join(path, "split_ids", f"{subset}_{split}.ids")
    with epath.Path(split_filename).open() as f:
      for line in f:
        json_path = os.path.join(path, subset, f"{line.rstrip()}.json")
        with epath.Path(json_path).open() as jf:
          d = json.load(jf)
          if style == "json":
            report = json.dumps(d[report_key])
          else:
            if subset == _SUBSET_CRS:
              report = _flatten_structure(
                  d[report_key], separator, 1, style == "html"
              )
            elif subset == _SUBSET_GAO:
              report = separator.join(
                  [
                      _flatten_structure(r, separator, 1, style == "html")
                      for r in d[report_key]
                  ]
              )
            else:
              raise ValueError("Unsupported subset.")

          if subset == _SUBSET_CRS:
            summary = separator.join(d[summary_key])
          elif subset == _SUBSET_GAO:
            summary = separator.join(
                sum([s["paragraphs"] for s in d[summary_key]], [])
            )
          else:
            raise ValueError("Unsupported subset.")
          results = {report_key: report, summary_key: ""}
          results.update(
              {
                  k: separator.join(d[k]) if isinstance(d[k], list) else d[k]
                  for k in self.builder_config.other_keys
              }
          )
          yield d[_ID_KEY], results


def _flatten_structure(
    d: Dict[str, Any], separator: str, depth: int, use_html: bool
) -> str:
  """Recursively flatten the structure."""
  texts = []
  title = d["section_title"]
  if use_html:
    texts.append(f"<h{depth}>{title}</h{depth}>")
  else:
    texts.append(title)
  texts.extend(d["paragraphs"])
  texts.extend(
      [
          _flatten_structure(s, separator, depth + 1, use_html)
          for s in d["subsections"]
      ]
  )
  return separator.join(texts)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Typing annotation utils."""

import typing
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf

_symbols_to_exclude = set(globals().keys())

T = TypeVar('T')

# Note: `TupleOrList` avoid abiguity from `Sequence` (`str` is `Sequence[str]`,
# `bytes` is `Sequence[int]`).
TupleOrList = Union[Tuple[T, ...], List[T]]
ListOrElem = Union[T, List[T]]

TreeDict = Union[T, Dict[str, 'TreeDict']]
Tree = Union[T, Any]
ListOrTreeOrElem = Union[T, TreeDict[T], List[T]]
NpArrayOrScalar = Union[bytes, float, int, np.ndarray, str]

if typing.TYPE_CHECKING:
  Tensor = Union[tf.Tensor, tf.SparseTensor, tf.RaggedTensor]
  TfdsDType = Union[np.dtype, Type[np.generic], tf.DType, tf.dtypes.DType]
else:
  Tensor = Any
  TfdsDType = Any

# Nested dict of tensor
TensorDict = TreeDict[Tensor]
NpArrayOrScalarDict = TreeDict[NpArrayOrScalar]

Dim = Optional[int]
Shape = TupleOrList[Dim]

JsonValue = Union[
    str,
    bool,
    int,
    float,
    None,
    List['JsonValue'],
    Dict[str, 'JsonValue'],  # pytype: disable=not-supported-yet
]
Json = Dict[str, JsonValue]

# Types for the tfrecord example construction.

Key = Union[int, str, bytes]
KeySerializedExample = Tuple[Key, bytes]  # `(key, serialized_proto)`


__all__ = sorted(
    k
    for k in globals()
    if k not in _symbols_to_exclude and not k.startswith('_')
)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A logger logging using absl.logging module."""

from typing import Optional, Union

from absl import logging
from tensorflow_datasets.core import decode
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core.logging import base_logger
from tensorflow_datasets.core.logging import call_metadata
from tensorflow_datasets.core.utils import read_config as tfds_read_config
from tensorflow_datasets.core.utils import type_utils

TreeDict = type_utils.TreeDict


class LoggingLogger(base_logger.Logger):
  """A logger that is based on the absl logging module."""

  def as_dataset(
      self,
      *,
      metadata: call_metadata.CallMetadata,
      name: str,
      config_name: Optional[str],
      version: str,
      data_path: str,
      split: Union[str, splits_lib.ReadInstruction],
      batch_size: Optional[int],
      shuffle_files: bool,
      read_config: tfds_read_config.ReadConfig,
      as_supervised: bool,
      decoders: Optional[TreeDict[decode.partial_decode.DecoderArg]],
  ):
    logging.info(
        "Constructing tf.data.Dataset %s for split %s, from %s",
        name,
        str(split),
        data_path,
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script that prints number of configs for a dataset."""

from absl import app
from absl import flags

import tensorflow_datasets as tfds

FLAGS = flags.FLAGS

flags.DEFINE_string("dataset", None, "DatasetBuilder to print num configs for")


def main(_):
  print(len(tfds.builder(FLAGS.dataset).BUILDER_CONFIGS))


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.utils.tf_utils."""

import numpy as np
import pytest
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core.utils import tf_utils


class TfUtilsTest(testing.TestCase):

  @testing.run_in_graph_and_eager_modes()
  def test_graph_runner(self):
    graph_runner = tf_utils.TFGraphRunner()

    output = graph_runner.run(tf.nn.relu, [1, 1, -1, -1, 1])
    self.assertAllEqual(output, [1, 1, 0, 0, 1])

    output = graph_runner.run(tf.nn.relu, [-1, -1, -1, 1, 1])
    self.assertAllEqual(output, [0, 0, 0, 1, 1])

    # Cache should have been re-used, so should only contains one GraphRun
    # Ideally there should be two separate @tf.eager.run_test_in_graph() and
    # @tf.eager.run_test_in_eager() to avoid logic on the test. But haven't
    # found it.
    if not tf.executing_eagerly():
      self.assertEqual(len(graph_runner._graph_run_cache), 1)
    else:
      self.assertEqual(len(graph_runner._graph_run_cache), 0)

    # Different signature (different shape), so new GraphRun created
    output = graph_runner.run(tf.nn.relu, [-1, 1, 1])
    self.assertAllEqual(output, [0, 1, 1])
    if not tf.executing_eagerly():
      self.assertEqual(len(graph_runner._graph_run_cache), 2)
    else:
      self.assertEqual(len(graph_runner._graph_run_cache), 0)


def test_shapes_are_compatible():
  assert tf_utils.shapes_are_compatible(
      {'a': {'b': (28, 28, 3)}},
      {'a': {'b': (None, None, 3)}},
  )
  assert not tf_utils.shapes_are_compatible(
      {'a': {'b': (28, 28, 3)}},
      {'a': {'b': (None, 27, 3)}},
  )


@pytest.mark.parametrize(
    ['tensor', 'np_shape', 'result'],
    [
        (
            tf.constant([28, 28, 3]),
            (None, None, 3),
            (tf.constant(28), tf.constant(28), 3),
        ),
        (np.array([28, 28, 3]), (None, None, 3), (28, 28, 3)),
    ],
)
def test_merge_shape(tensor, np_shape, result):
  assert tf_utils.merge_shape(tensor, np_shape) == result


@pytest.mark.parametrize(
    ['shape1', 'shape2'],
    [
        (None, None),
        ((None,), (None,)),
        ((1,), (1,)),
        ((None,), (1,)),
        ((1,), (None,)),
        ((1, 2), (1, 2)),
        ((1, 2), (None, None)),
        ((1, None), (None, None)),
        ((None, 2), (None, None)),
        ((None, None), (1, 2)),
        ((None, None), (1, None)),
        ((None, None), (None, 2)),
    ],
)
def test_assert_shapes_match(shape1, shape2):
  try:
    tf_utils.assert_shape_match(shape1, shape2)
  except ValueError as exception:
    raise Exception(f'test should fail for {shape1}/{shape2}') from exception


@pytest.mark.parametrize(
    ['shape1', 'shape2'],
    [
        ((1, 2), (3, 2)),
        ((1, 2), (1, 3)),
    ],
)
def test_assert_must_have_the_same_dimension(shape1, shape2):
  with pytest.raises(ValueError, match='are incompatible'):
    tf_utils.assert_shape_match(shape1, shape2)


@pytest.mark.parametrize(
    ['shape1', 'shape2'],
    [
        (None, (1,)),
        (None, (1, 2)),
        ((1,), None),
        ((1, 2), None),
        ((1,), (1, 2)),
        ((1, 2), (1,)),
    ],
)
def test_assert_must_have_the_same_rank(shape1, shape2):
  with pytest.raises(ValueError, match='must have the same rank'):
    tf_utils.assert_shape_match(shape1, shape2)


if __name__ == '__main__':
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Update checksums to contain filename."""

from concurrent import futures
import dataclasses
import itertools
import typing
from typing import Dict, NewType, Optional
import urllib

from absl import app
from absl import logging
import requests
import tensorflow_datasets as tfds
from tensorflow_datasets.core.download import checksums
import tqdm

Url = NewType('Url', str)
Filename = NewType('Filename', Optional[str])

# pylint: disable=logging-format-interpolation


def _collect_path_to_url_infos() -> (
    Dict[tfds.core.Path, Dict[Url, checksums.UrlInfo]]
):
  """Collect checksums paths to url_infos."""
  # Collect legacy checksums paths
  url_info_paths = list(checksums._checksum_paths().values())  # pylint: disable=protected-access

  # Collect dataset-as-folder checksums path
  for name in tfds.list_builders(with_community_datasets=False):
    url_info_path = tfds.builder_cls(name)._checksums_path  # pylint: disable=protected-access
    if url_info_path.exists():
      url_info_paths.append(url_info_path)

  url_info_paths = [tfds.core.utils.to_write_path(p) for p in url_info_paths]
  return {
      path: typing.cast(
          Dict[Url, checksums.UrlInfo], checksums.load_url_infos(path)
      )
      for path in url_info_paths
  }


def _request_filename(
    url: Url,
    url_info: checksums.UrlInfo,
    tqdm_bar: tqdm.tqdm,
) -> Filename:
  """Get filename of dataset at `url`."""
  if url_info.filename:
    # tqdm_bar.update(1)
    return Filename(url_info.filename)
  # Uses `_open_url` for drive urls compatibility
  downloader = tfds.core.download.downloader
  try:
    with downloader._open_url(url, verify=True, timeout=10.0) as (response, _):  # pylint: disable=protected-access
      filename = downloader._get_filename(response)  # pylint: disable=protected-access
  except (
      requests.exceptions.RequestException,
      urllib.error.URLError,
      downloader.DownloadError,
  ) as e:
    tqdm.tqdm.write(f'Error for {url}: {e}')
    filename = None
  tqdm_bar.update(1)
  # logging.info(f'{url} -> {filename}')
  return Filename(filename)


def _update_url_info(
    url: Url,
    url_info: checksums.UrlInfo,
    new_filename: Filename,
) -> checksums.UrlInfo:
  """Update the `UrlInfo` with the filename."""
  old_filename = url_info.filename
  if new_filename is None:
    return url_info
  if old_filename and old_filename != new_filename:
    tqdm.tqdm.write(
        f'Filename for {url} already exist. Updating `{old_filename}` to '
        f'`{new_filename}`'
    )
  return dataclasses.replace(url_info, filename=new_filename)


def main(_):
  logging.info('Load all UrlInfo.')
  path_to_url_infos = _collect_path_to_url_infos()

  # Remove duplicate urls (merge all Dict[Url, UrlInfo] together)
  all_url_infos = dict(
      itertools.chain.from_iterable(
          d.items() for d in path_to_url_infos.values()
      )
  )

  logging.info('Start fetching filenames.')
  with futures.ThreadPoolExecutor(max_workers=100) as executor:
    # Query all filenames in parallel
    iter_all_url_infos = tqdm.tqdm(all_url_infos.items(), desc='Urls sent')
    received_tqdm = tqdm.tqdm(
        desc='Filenames received', total=len(all_url_infos)
    )
    urls_to_filename = {
        url: executor.submit(_request_filename, url, url_info, received_tqdm)
        for url, url_info in iter_all_url_infos
    }
    # Update and save the new UrlInfo
    for path, url_infos in tqdm.tqdm(
        path_to_url_infos.items(), desc='Saving filenames'
    ):
      # Update the UrlInfo with the new filenames
      url_infos = {
          url: _update_url_info(url, url_info, urls_to_filename[url].result())
          for url, url_info in url_infos.items()
      }
      checksums.save_url_infos(path, url_infos)
  received_tqdm.close()
  tqdm.tqdm.write('Url infos updated.')


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding=utf-8
"""TextEncoders convert between text and integers."""
from __future__ import unicode_literals

import abc
import hashlib
import json
import re

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


def _re_compile(pattern):
  return re.compile(pattern, flags=re.UNICODE)  # pytype: disable=wrong-keyword-args


NUM_BYTES = 2**8
ALPHANUM_REGEX = _re_compile(r"\W+")
ALL_REGEX = _re_compile(r"(\W+)")



class TextEncoderConfig:
  """Configuration for `tfds.features.Text`."""

  def __init__(
      self, encoder=None, encoder_cls=None, vocab_size=None, name=None
  ):
    if encoder:
      if encoder_cls or vocab_size:
        raise ValueError(
            "If encoder is provided, encoder_cls and vocab_size must be None"
        )
      encoder_cls = type(encoder)
      vocab_size = encoder.vocab_size
    else:
      if encoder_cls is ByteTextEncoder:
        encoder = encoder_cls()

    self.encoder = encoder
    self.encoder_cls = encoder_cls
    self.vocab_size = vocab_size
    self.name = name


class TextEncoder(metaclass=abc.ABCMeta):
  """Abstract base class for converting between text and integers.

  **A note on padding**:

    Because text data is typically variable length and nearly always requires
    padding during training, ID 0 is always reserved for padding. To accommodate
    this, all `TextEncoder`s behave in certain ways:

    * `encode`: never returns id 0 (all ids are 1+)
    * `decode`: drops 0 in the input ids
    * `vocab_size`: includes ID 0

    New subclasses should be careful to match this behavior.
  """

  @abc.abstractmethod
  def encode(self, s):
    """Encodes text into a list of integers."""
    raise NotImplementedError

  @abc.abstractmethod
  def decode(self, ids):
    """Decodes a list of integers into text."""
    raise NotImplementedError

  @abc.abstractproperty
  def vocab_size(self):
    """Size of the vocabulary. Decode produces ints [1, vocab_size)."""
    raise NotImplementedError

  @abc.abstractmethod
  def save_to_file(self, filename_prefix):
    """Store to file. Inverse of load_from_file."""
    raise NotImplementedError

  @classmethod
  @abc.abstractmethod
  def load_from_file(cls, filename_prefix):  # pylint: disable=no-self-argument
    """Load from file. Inverse of save_to_file."""
    raise NotImplementedError

  @classmethod
  def _write_lines_to_file(cls, filename, lines, metadata_dict=None):
    """Writes lines to file prepended by header and metadata."""
    write_lines_to_file(cls.__name__, filename, lines, metadata_dict)

  @classmethod
  def _read_lines_from_file(cls, filename):
    return read_lines_from_file(cls.__name__, filename)

  def __repr__(self):
    return "<%s vocab_size=%d>" % (type(self).__name__, self.vocab_size)


class ByteTextEncoder(TextEncoder):
  """Byte-encodes text."""

  def __init__(self, additional_tokens=None):
    """Constructs ByteTextEncoder.

    Args:
      additional_tokens: `list<str>`, list of additional tokens. These will be
        assigned vocab ids `[1, 1+len(additional_tokens)]`. Useful for things
        like "end-of-string" tokens (e.g. "<EOS>").
    """
    self._additional_tokens, self._additional_tokens_re = (
        _prepare_reserved_tokens(additional_tokens)
    )
    # Note that internally everything is 0-indexed. Padding is dealt with at the
    # end of encode and the beginning of decode.
    self._additional_token_to_id = dict(
        zip(self._additional_tokens, range(len(self._additional_tokens)))
    )

  def encode(self, s):
    if not self.additional_tokens:
      return pad_incr(list(bytearray(tf.compat.as_bytes(s))))

    # Handle additional tokens
    s = tf.compat.as_text(s)
    ids = []
    for substr in self._additional_tokens_re.split(s):
      if not substr:
        continue
      tok_id = self._additional_token_to_id.get(substr)
      if tok_id is None:
        offset = len(self.additional_tokens)
        tok_ids = [
            i + offset for i in list(bytearray(tf.compat.as_bytes(substr)))
        ]
      else:
        tok_ids = [tok_id]
      ids.extend(tok_ids)

    return pad_incr(ids)

  def decode(self, ids):
    ids = pad_decr(ids)
    if not self.additional_tokens:
      return tf.compat.as_text(bytes(bytearray(ids)))

    # Handle additional tokens
    # First pass picks out the additional tokens
    tmp_decoded = []
    for byte_id in ids:
      is_additional_token = byte_id < len(self.additional_tokens)
      if is_additional_token:
        tmp_decoded.append(self.additional_tokens[byte_id])
      else:
        # Leave these as ints so that we can contiguously decode bytes
        # afterwards
        tmp_decoded.append(byte_id - len(self.additional_tokens))

    # Second pass to decode contiguous bytes
    strs = []
    i = 0
    while i < len(tmp_decoded):
      el = tmp_decoded[i]
      if isinstance(el, str):
        strs.append(el)
        i += 1
      else:
        # Decode contiguous bytes
        byte_ids = []
        while i < len(tmp_decoded):
          b = tmp_decoded[i]
          if isinstance(b, int):
            byte_ids.append(b)
            i += 1
          else:
            break
        strs.append(bytes(bytearray(byte_ids)).decode("utf-8", "replace"))
    return "".join(strs)

  @property
  def vocab_size(self):
    # Plus 1 for pad
    return len(self.additional_tokens) + NUM_BYTES + 1

  @property
  def additional_tokens(self):
    return self._additional_tokens

  @classmethod
  def _filename(cls, filename_prefix):
    return filename_prefix + ".bytes"

  def save_to_file(self, filename_prefix):
    self._write_lines_to_file(
        self._filename(filename_prefix), self.additional_tokens
    )

  @classmethod
  def load_from_file(cls, filename_prefix):
    lines, _ = cls._read_lines_from_file(cls._filename(filename_prefix))
    return cls(additional_tokens=lines)


class TokenTextEncoder(TextEncoder):
  r"""TextEncoder backed by a list of tokens.

  Tokenization splits on (and drops) non-alphanumeric characters with
  regex "\W+".
  """

  def __init__(
      self,
      vocab_list,
      oov_buckets=1,
      oov_token="UNK",
      lowercase=False,
      tokenizer=None,
      strip_vocab=True,
      decode_token_separator=" ",
  ):
    """Constructs a TokenTextEncoder.

    To load from a file saved with `TokenTextEncoder.save_to_file`, use
    `TokenTextEncoder.load_from_file`.

    Args:
      vocab_list: `list<str>`, list of tokens.
      oov_buckets: `int`, the number of `int`s to reserve for OOV hash buckets.
        Tokens that are OOV will be hash-modded into a OOV bucket in `encode`.
      oov_token: `str`, the string to use for OOV ids in `decode`.
      lowercase: `bool`, whether to make all text and tokens lowercase.
      tokenizer: `Tokenizer`, responsible for converting incoming text into a
        list of tokens.
      strip_vocab: `bool`, whether to strip whitespace from the beginning and
        end of elements of `vocab_list`.
      decode_token_separator: `str`, the string used to separate tokens when
        decoding.
    """
    self._vocab_list = [tf.compat.as_text(el) for el in vocab_list]
    if strip_vocab:
      self._vocab_list = [el.strip() for el in self._vocab_list]
    self._lowercase = lowercase
    if self._lowercase:
      self._vocab_list = [t.lower() for t in self._vocab_list]
    # Note that internally everything is 0-indexed. Padding is dealt with at the
    # end of encode and the beginning of decode.
    self._token_to_id = dict(
        zip(self._vocab_list, range(len(self._vocab_list)))
    )
    self._oov_buckets = oov_buckets
    self._oov_token = tf.compat.as_text(oov_token)

    # Reserved tokens are all tokens that are mixed alphanum and non-alphanum.
    reserved_tokens = [t for t in self._vocab_list if is_mixed_alphanum(t)]
    self._tokenizer = tokenizer or Tokenizer(reserved_tokens=reserved_tokens)
    self._user_defined_tokenizer = tokenizer

    self._decode_token_separator = decode_token_separator

  def encode(self, s):
    s = tf.compat.as_text(s)
    if self.lowercase:
      s = s.lower()
    ids = []
    for token in self._tokenizer.tokenize(s):
      int_id = self._token_to_id.get(token, -1)
      if int_id < 0:
        int_id = self._oov_bucket(token)
        if int_id is None:
          raise ValueError("Out of vocabulary token %s" % token)
      ids.append(int_id)

    # Increment for pad id 0
    return pad_incr(ids)

  def decode(self, ids):
    ids = pad_decr(ids)

    tokens = []
    for int_id in ids:
      if int_id < len(self._vocab_list):
        tokens.append(self._vocab_list[int_id])
      else:
        tokens.append(self._oov_token)
    return self._decode_token_separator.join(tokens)

  @property
  def vocab_size(self):
    # Plus 1 for pad
    return len(self._vocab_list) + self._oov_buckets + 1

  @property
  def tokens(self):
    return list(self._vocab_list)

  @property
  def oov_token(self):
    return self._oov_token

  @property
  def lowercase(self):
    return self._lowercase

  @property
  def tokenizer(self):
    return self._tokenizer

  def _oov_bucket(self, token):
    if self._oov_buckets <= 0:
      return None
    if self._oov_buckets == 1:
      return len(self._vocab_list)
    hash_val = int(hashlib.md5(tf.compat.as_bytes(token)).hexdigest(), 16)
    return len(self._vocab_list) + hash_val % self._oov_buckets

  @classmethod
  def _filename(cls, filename_prefix):
    return filename_prefix + ".tokens"

  def save_to_file(self, filename_prefix):
    filename = self._filename(filename_prefix)
    kwargs = {
        "oov_buckets": self._oov_buckets,
        "lowercase": self._lowercase,
        "oov_token": self._oov_token,
    }
    if self._user_defined_tokenizer is not None:
      self._tokenizer.save_to_file(filename)
      kwargs["has_tokenizer"] = True
    self._write_lines_to_file(filename, self._vocab_list, kwargs)

  @classmethod
  def load_from_file(cls, filename_prefix):
    filename = cls._filename(filename_prefix)
    vocab_lines, kwargs = cls._read_lines_from_file(filename)
    has_tokenizer = kwargs.pop("has_tokenizer", False)
    if has_tokenizer:
      kwargs["tokenizer"] = Tokenizer.load_from_file(filename)
    return cls(vocab_list=vocab_lines, **kwargs)


class Tokenizer(object):
  """Splits a string into tokens, and joins them back."""

  def __init__(self, alphanum_only=True, reserved_tokens=None):
    """Constructs a Tokenizer.

    Note that the Tokenizer is invertible if `alphanum_only=False`.
    i.e. `s == t.join(t.tokenize(s))`.

    Args:
      alphanum_only: `bool`, if `True`, only parse out alphanumeric tokens
        (non-alphanumeric characters are dropped); otherwise, keep all
        characters (individual tokens will still be either all alphanumeric or
        all non-alphanumeric).
      reserved_tokens: `list<str>`, a list of strings that, if any are in `s`,
        will be preserved as whole tokens, even if they contain mixed
        alphanumeric/non-alphanumeric characters.
    """
    self._alphanum_only = alphanum_only
    reserved_tokens, self._reserved_tokens_re = _prepare_reserved_tokens(
        reserved_tokens
    )
    self._reserved_tokens = set(reserved_tokens)

  @property
  def alphanum_only(self):
    return self._alphanum_only

  @property
  def reserved_tokens(self):
    return self._reserved_tokens

  def tokenize(self, s):
    """Splits a string into tokens."""
    s = tf.compat.as_text(s)

    if self.reserved_tokens:
      # First split out the reserved tokens
      substrs = self._reserved_tokens_re.split(s)
    else:
      substrs = [s]

    toks = []
    for substr in substrs:
      if substr in self.reserved_tokens:
        toks.append(substr)
      elif self._alphanum_only:
        toks.extend(ALPHANUM_REGEX.split(substr))
      else:
        toks.extend(ALL_REGEX.split(substr))

    # Filter out empty strings
    toks = [t for t in toks if t]
    return toks

  def join(self, tokens):
    """Joins tokens into a string."""
    if self._alphanum_only:
      return " ".join(tokens)
    else:
      # Fully invertible
      return "".join(tokens)

  @classmethod
  def _filename(cls, filename_prefix):
    return filename_prefix + ".tokenizer"

  def save_to_file(self, filename_prefix):
    filename = self._filename(filename_prefix)
    kwargs = {
        "reserved_tokens": list(self._reserved_tokens),
        "alphanum_only": self._alphanum_only,
    }
    write_lines_to_file(type(self).__name__, filename, [], kwargs)

  @classmethod
  def load_from_file(cls, filename_prefix):
    filename = cls._filename(filename_prefix)
    _, kwargs = read_lines_from_file(cls.__name__, filename)
    return cls(**kwargs)


def pad_decr(ids):
  """Strip ID 0 and decrement ids by 1."""
  if len(ids) < 1:
    return list(ids)
  if not any(ids):
    return []  # all padding.
  idx = -1
  while not ids[idx]:
    idx -= 1
  if idx == -1:
    ids = ids  # pylint: disable=self-assigning-variable
  else:
    ids = ids[: idx + 1]
  return [i - 1 for i in ids]


def pad_incr(ids):
  """Add 1 to ids to account for pad."""
  return [i + 1 for i in ids]


def _prepare_reserved_tokens(reserved_tokens):
  """Prepare reserved tokens and a regex for splitting them out of strings."""
  reserved_tokens = [tf.compat.as_text(tok) for tok in reserved_tokens or []]
  dups = _find_duplicates(reserved_tokens)
  if dups:
    raise ValueError("Duplicates found in tokens: %s" % dups)
  reserved_tokens_re = _make_reserved_tokens_re(reserved_tokens)
  return reserved_tokens, reserved_tokens_re


def _re_escape(s):
  """Escape regex control characters."""
  escaped = re.sub(r"[(){}\[\].*?|^$\\+-]", r"\\\g<0>", s)
  return escaped


def _make_reserved_tokens_re(reserved_tokens):
  """Constructs compiled regex to parse out reserved tokens."""
  if not reserved_tokens:
    return None
  escaped_tokens = [_re_escape(rt) for rt in reserved_tokens]
  pattern = "(%s)" % "|".join(escaped_tokens)
  reserved_tokens_re = _re_compile(pattern)
  return reserved_tokens_re


def _find_duplicates(els):
  seen = set()
  dups = []
  for x in els:
    if x in seen:
      dups.append(x)
    else:
      seen.add(x)
  return dups


def is_mixed_alphanum(token):
  return len([s for s in ALL_REGEX.split(token) if s]) > 1


_HEADER_PREFIX = "### "
_METADATA_PREFIX = "### Metadata: "


def write_lines_to_file(cls_name, filename, lines, metadata_dict):
  """Writes lines to file prepended by header and metadata."""
  metadata_dict = metadata_dict or {}
  header_line = "%s%s" % (_HEADER_PREFIX, cls_name)
  metadata_line = "%s%s" % (
      _METADATA_PREFIX,
      json.dumps(metadata_dict, sort_keys=True),
  )
  with tf.io.gfile.GFile(filename, "wb") as f:
    for line in [header_line, metadata_line]:
      f.write(tf.compat.as_bytes(line))
      f.write(tf.compat.as_bytes("\n"))
    if lines:
      f.write(tf.compat.as_bytes("\n".join(lines)))
      f.write(tf.compat.as_bytes("\n"))


def read_lines_from_file(cls_name, filename):
  """Read lines from file, parsing out header and metadata."""
  with tf.io.gfile.GFile(filename, "rb") as f:
    lines = [tf.compat.as_text(line)[:-1] for line in f]
  header_line = "%s%s" % (_HEADER_PREFIX, cls_name)
  if lines[0] != header_line:
    raise ValueError(
        "File {fname} does not seem to have been created from "
        "{name}.save_to_file.".format(fname=filename, name=cls_name)
    )
  metadata_dict = json.loads(lines[1][len(_METADATA_PREFIX) :])
  return lines[2:], metadata_dict



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Script which parse registered repositories and save datasets found.

"""

import json
from typing import List

from absl import app
import tensorflow_datasets as tfds
import toml
import tqdm

# Community datasets namespaces and code location from where the datasets
# package index will be constructed.
_IN_PATH = tfds.core.tfds_path() / 'community-datasets.toml'
# Community datasets package indexes which will be updated.
_OUT_PATH = tfds.core.utils.gcs_utils.GCS_COMMUNITY_INDEX_PATH

DatasetSource = tfds.core.community.dataset_sources.DatasetSource
DatasetPackage = tfds.core.community.register_package.DatasetPackage


def main(_):
  export_community_datasets(in_path=_IN_PATH, out_path=_OUT_PATH)  # pytype: disable=wrong-arg-types


def export_community_datasets(
    in_path: tfds.core.Path,
    out_path: tfds.core.Path,
) -> None:
  """Exports community datasets.

  Args:
    in_path: Config path containing the namespaces and dataset lookup
      instructions.
    out_path: File containing all detected datasets. Detected dataset will be
      saved to this file. Previous content is erased.
  """
  ds_packages = _find_community_ds_packages(in_path)
  _save_community_ds_packages(out_path, ds_packages)


def _find_community_ds_packages(
    config_path: tfds.core.Path,
) -> List[DatasetPackage]:
  """Find all namepaces/dataset from the config.

  Config should contain the instructions in the following format:

  ```
  [Namespace]
  <namespace0> = 'github://<owner0>/<github_repo0>/tree/<path/to/dataset/dir>'
  <namespace1> = 'gs://<bucket>/<datasets>/'
  ```

  Args:
    config_path: Path to the config file containing lookup instructions.

  Returns:
    ds_packages: list of all found datasets.
  """
  config = toml.load(config_path)

  all_packages = []
  for namespace, src_code_path in tqdm.tqdm(config['Namespaces'].items()):
    tqdm.tqdm.write(f'Searching datasets for {namespace}: {src_code_path}')
    for (
        pkg
    ) in tfds.core.community.register_package.list_ds_packages_for_namespace(
        namespace=namespace, path=tfds.core.Path(src_code_path)
    ):
      tqdm.tqdm.write(str(pkg.name))
      all_packages.append(pkg)

  return sorted(all_packages, key=lambda package: package.name)


def _save_community_ds_packages(
    file_path: tfds.core.Path, ds_packages: List[DatasetPackage]
) -> None:
  """Save all loaded datasets in the package index.

  Saved file will have the following `.jsonl` format:

  ```jsonl
  {'name': 'namespace0:dataset0', 'source': 'github://...'}
  {'name': 'namespace0:dataset0', 'source': 'github://...'}
  ...
  ```

  Args:
    file_path: `.jsonl` destination to which save the dataset
    ds_packages: Dataset paths to save
  """
  pkg_json = [json.dumps(pkg.to_json()) for pkg in ds_packages]
  file_path.write_text('\n'.join(pkg_json) + '\n')


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""domainnet dataset."""

from tensorflow_datasets.image_classification.domainnet import domainnet
import tensorflow_datasets.public_api as tfds


class DomainnetTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for domainnet dataset."""

  DATASET_CLASS = domainnet.Domainnet
  BUILDER_CONFIG_NAMES_TO_TEST = ['real']
  SPLITS = {
      'train': 2,  # Number of fake train example
      'test': 1,  # Number of fake train example
  }
  DL_DOWNLOAD_RESULT = {'train': 'real_train.txt', 'test': 'real_test.txt'}
  DL_EXTRACT_RESULT = 'img'


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VoxForge dataset."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import datetime
import io
import os
import textwrap

import numpy as np
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@article{maclean2018voxforge,
  title={Voxforge},
  author={MacLean, Ken},
  journal={Ken MacLean.[Online]. Available: http://www.voxforge.org/home.[Acedido em 2012]},
  year={2018}
}
"""

_LAST_DATE = datetime.date(2020, 1, 1)

_DESCRIPTION = """
VoxForge is a language classification dataset. It consists of user submitted
audio clips submitted to the website. In this release, data from 6 languages
is collected - English, Spanish, French, German, Russian, and Italian.
Since the website is constantly updated, and for the sake of reproducibility,
this release contains only recordings submitted prior to {}.
The samples are splitted between train, validation and testing so that samples
from each speaker belongs to exactly one split.
""".format(
    _LAST_DATE.isoformat()
)

_HOMEPAGE = 'http://www.voxforge.org/'

_SAMPLE_RATE = 16000

_URLS_LIST_FILE = 'https://storage.googleapis.com/tfds-data/downloads/voxforge/voxforge_urls.txt'

LABELS = ['de', 'en', 'es', 'fr', 'it', 'ru']


def _compute_split_boundaries(split_probs, n_items):
  """Computes boundary indices for each of the splits in split_probs.

  Args:
    split_probs: List of (split_name, prob), e.g. [('train', 0.6), ('dev', 0.2),
      ('test', 0.2)]
    n_items: Number of items we want to split.

  Returns:
    The item indices of boundaries between different splits. For the above
    example and n_items=100, these will be
    [('train', 0, 60), ('dev', 60, 80), ('test', 80, 100)].
  """
  if len(split_probs) > n_items:
    raise ValueError(
        'Not enough items for the splits. There are {splits} '
        'splits while there are only {items} items'.format(
            splits=len(split_probs), items=n_items
        )
    )
  total_probs = sum(p for name, p in split_probs)
  if abs(1 - total_probs) > 1e-8:
    raise ValueError('Probs should sum up to 1. probs={}'.format(split_probs))
  split_boundaries = []
  sum_p = 0.0
  for name, p in split_probs:
    prev = sum_p
    sum_p += p
    split_boundaries.append((name, int(prev * n_items), int(sum_p * n_items)))

  # Guard against rounding errors.
  split_boundaries[-1] = (
      split_boundaries[-1][0],
      split_boundaries[-1][1],
      n_items,
  )

  return split_boundaries


def _wav_obj_to_samples(wav_obj):
  """Read a `tarfile.ExFileObject`."""
  sample_rate, samples = tfds.core.lazy_imports.scipy.io.wavfile.read(
      io.BytesIO(wav_obj.read())
  )
  return samples, sample_rate


def _get_inter_splits_by_group(items_and_groups, split_probs, split_number):
  """Split items to train/dev/test, so all items in group go into same split.

  Each group contains all the samples from the same speaker ID. The samples are
  splitted between train, validation and testing so that samples from each
  speaker belongs to exactly one split.

  Args:
    items_and_groups: Sequence of (item_id, group_id) pairs.
    split_probs: List of (split_name, prob), e.g. [('train', 0.6), ('dev', 0.2),
      ('test', 0.2)]
    split_number: Generated splits should change with split_number.

  Returns:
    Dictionary that looks like {split name -> set(ids)}.
  """
  groups = sorted(set(group_id for item_id, group_id in items_and_groups))
  rng = np.random.RandomState(split_number)
  rng.shuffle(groups)

  split_boundaries = _compute_split_boundaries(split_probs, len(groups))
  group_id_to_split = {}
  for split_name, i_start, i_end in split_boundaries:
    for i in range(i_start, i_end):
      group_id_to_split[groups[i]] = split_name

  split_to_ids = collections.defaultdict(set)
  for item_id, group_id in items_and_groups:
    split = group_id_to_split[group_id]
    split_to_ids[split].add(item_id)

  return split_to_ids


class Voxforge(tfds.core.BeamBasedBuilder):
  """A Language classification dataset based on the VoxForge website."""

  VERSION = tfds.core.Version('1.0.0')

  MANUAL_DOWNLOAD_INSTRUCTIONS = textwrap.dedent(
      """
  VoxForge requires manual download of the audio archives. The complete list of
  archives can be found in {}. It can be downloaded using the following command:
  wget -i voxforge_urls.txt -x
  Note that downloading and building the dataset locally requires ~100GB disk
  space (but only ~60GB will be used permanently).
  """.format(
          _URLS_LIST_FILE
      )
  )

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'audio': tfds.features.Audio(
                file_format='wav', sample_rate=_SAMPLE_RATE
            ),
            'label': tfds.features.ClassLabel(names=LABELS),
            'speaker_id': np.str_,
        }),
        supervised_keys=('audio', 'label'),
        homepage=_HOMEPAGE,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    urls_list = dl_manager.download({'urls_list': _URLS_LIST_FILE})
    archive_urls = []
    with tf.io.gfile.GFile(urls_list['urls_list']) as f:
      for line in f:
        archive_url = line.strip().replace('"', '').replace("'", '')
        archive_path = os.path.join(dl_manager.manual_dir, archive_url)
        if not tf.io.gfile.exists(archive_path):
          raise AssertionError(
              'VoxForge requires manual download. Path {} is missing'.format(
                  archive_path
              )
          )
        archive_urls.append(archive_path)

    archives_and_speaker_ids = []
    for archive_url in archive_urls:
      _, archive_name = os.path.split(archive_url)
      speaker_id = archive_name.split('-')[0]
      archives_and_speaker_ids.append((archive_url, speaker_id))

    split_probs = [('train', 0.7), ('validation', 0.1), ('test', 0.2)]
    splits = _get_inter_splits_by_group(
        archives_and_speaker_ids, split_probs, 0
    )

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={
                'file_names': splits['train'],
                'dl_manager': dl_manager,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={
                'file_names': splits['validation'],
                'dl_manager': dl_manager,
            },
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={'file_names': splits['test'], 'dl_manager': dl_manager},
        ),
    ]

  def _build_pcollection(self, pipeline, file_names, dl_manager):
    """Build Pcollection of examples."""
    beam = tfds.core.lazy_imports.apache_beam
    return (
        pipeline
        | beam.Create(file_names)
        | beam.FlatMap(_generate_examples, dl_manager=dl_manager)
    )


def _generate_examples(fname, dl_manager):
  """Yields examples."""
  folder, archive_name = os.path.split(fname)
  speaker_id = archive_name.split('-')[0]
  label_idx = folder.index('/Trunk') - 2
  label = folder[label_idx : label_idx + 2]
  iter_archive = dl_manager.iter_archive(fname)
  for wav_path, wav_obj in iter_archive:
    if not wav_path.endswith('.wav'):
      continue
    _, wav_name = os.path.split(wav_path)
    key = '{}_{}_{}'.format(label, archive_name, wav_name[: -len('.wav')])
    samples, sample_rate = _wav_obj_to_samples(wav_obj)
    if sample_rate != _SAMPLE_RATE:
      raise ValueError(
          f'Data sample rate was {sample_rate}, but must be {_SAMPLE_RATE}'
      )
    yield key, {'audio': samples, 'label': label, 'speaker_id': speaker_id}



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TFDS logging module."""
import abc
import atexit
import collections
import functools
import threading
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar

from absl import flags
from tensorflow_datasets.core.logging import base_logger
from tensorflow_datasets.core.logging import call_metadata
from tensorflow_datasets.core.logging import logging_logger
import wrapt


_T = TypeVar("_T")
_Decorator = Callable[[_T], _T]
_LoggerMethod = Callable[..., None]


_registered_loggers: Optional[List[base_logger.Logger]] = None

_import_operations: List[Tuple[call_metadata.CallMetadata, int, int]] = []
_import_operations_lock = threading.Lock()

_thread_id_to_builder_init_count = collections.Counter()


def _init_registered_loggers() -> List[base_logger.Logger]:
  """Initializes the registered loggers if they are not set yet."""
  global _registered_loggers
  if _registered_loggers is None:
    _registered_loggers = [
        logging_logger.LoggingLogger(),
    ]
  return _registered_loggers


def _log_import_operation():
  """Log import operations (most of time maximum one), if any."""
  with _import_operations_lock:
    for metadata, import_time_tf, import_time_builders in _import_operations:
      for logger in _init_registered_loggers():
        logger.tfds_import(
            metadata=metadata,
            import_time_ms_tensorflow=import_time_tf,
            import_time_ms_dataset_builders=import_time_builders,
        )
    _import_operations.clear()


def _get_registered_loggers() -> List[base_logger.Logger]:
  _log_import_operation()
  return _init_registered_loggers()


@atexit.register
def _at_exit():
  """Log import operation if needed, calls `process_ends` on loggers."""
  for logger in _get_registered_loggers():
    logger.process_ends()


class _FunctionDecorator(abc.ABC):
  """Base class for a TFDS function decorator.

  When a decorated function is called, the method with the same name as this
  class is called on the registered loggers.
  """

  def _start_call(self) -> call_metadata.CallMetadata:
    """Initializes call metadata."""
    return call_metadata.CallMetadata()

  def _fill_logger_method_kwargs(
      self,
      logger_method: _LoggerMethod,
      metadata: call_metadata.CallMetadata,
      instance: Any,
      args: Any,
      kwargs: Any,
  ) -> _LoggerMethod:
    """Fills kwargs for the corresponding logger method.

    Args:
      logger_method: logger method with the same name as this class.
      metadata: call metadata.
      instance: the object to which the wrapped function was bound when it was
        called.
      args: args captured during the function call.
      kwargs: kwargs captured during the function call.

    Returns:
      partial logger method with filled kwargs.
    """
    del instance, args, kwargs
    return functools.partial(logger_method, metadata=metadata)

  @abc.abstractmethod
  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    """Calls the logger method with appropriate `args` and `kwargs`.

    Child classes need to implement how `args` and `kwargs` are passed from the
    function to the corresponding logger method.

    Args:
      logger_method: logger method with the same name as this class.
      args: args captured during the function call.
      kwargs: kwargs captured during the function call.
    """
    pass

  def _finish_call(
      self,
      metadata: call_metadata.CallMetadata,
      instance: Any,
      args: Any,
      kwargs: Any,
  ):
    """Completes wrapped function call.

    Metadata marks the end and the corresponding method is called on the
    registered loggers.

    Args:
      metadata: call metadata.
      instance: the object to which the wrapped function was bound when it was
        called.
      args: args captured during the function call.
      kwargs: kwargs captured during the function call.
    """
    metadata.mark_end()
    for logger in _get_registered_loggers():
      method_name = self.__class__.__name__
      logger_method = getattr(logger, method_name)
      logger_method = self._fill_logger_method_kwargs(
          logger_method, metadata, instance, args, kwargs
      )
      self._call_logger_method(logger_method, args, kwargs)

  @wrapt.decorator
  def __call__(self, function, instance, args, kwargs):
    """Decorator to call the corresponding method on registered loggers."""
    metadata = self._start_call()
    try:
      return function(*args, **kwargs)
    except Exception:
      metadata.mark_error()
      raise
    finally:
      self._finish_call(metadata, instance, args, kwargs)


class _DsbuilderMethodDecorator(_FunctionDecorator):
  """Base class for a dataset builder method decorator."""

  # Flag. If `True` then the wrapped dsbuilder method is a property.
  IS_PROPERTY: bool = False

  @staticmethod
  def _get_info(dsbuilder: Any) -> Tuple[str, str, str, str]:
    """Gets information about the builder.

    Args:
      dsbuilder: the builder instance to which the wrapped function was bound
        when it was called.

    Returns:
      builder_name: name of the builder.
      config: name of the builder config.
      version: version of the builder.
      data_path: path to the builder data directory.
    """
    config_name = (
        dsbuilder.builder_config.name if dsbuilder.builder_config else ""
    )
    data_path = dsbuilder.data_dir
    return dsbuilder.name, config_name, str(dsbuilder.version), data_path

  def _fill_logger_method_kwargs(
      self,
      logger_method: _LoggerMethod,
      metadata: call_metadata.CallMetadata,
      instance: Any,
      args: Any,
      kwargs: Any,
  ) -> _LoggerMethod:
    """Fills kwargs for the corresponding logger method.

    Args:
      logger_method: logger method with the same name as this class.
      metadata: call metadata.
      instance: the object to which the wrapped function was bound when it was
        called.
      args: args captured during the function call.
      kwargs: kwargs captured during the function call.

    Returns:
      partial logger method with filled kwargs.
    """
    del kwargs
    if self.IS_PROPERTY:
      dsbuilder = args[0]  # Because property decorator applied first.
    else:
      dsbuilder = instance
    name, config_name, version, data_path = self._get_info(dsbuilder)

    return functools.partial(
        logger_method,
        metadata=metadata,
        name=name,
        config_name=config_name,
        version=version,
        data_path=data_path,
    )




def register(logger: base_logger.Logger) -> None:
  """Register an additional logger within TFDS.

  Registered loggers are called on TFDS events, synchronously, from the same
  thread the call was made, in sequence, and in registration order. Exceptions
  are *NOT* caught.

  Args:
    logger: the logger to register.
  """
  _init_registered_loggers().append(logger)


def tfds_import(
    *,
    metadata: call_metadata.CallMetadata,
    import_time_ms_tensorflow: int,
    import_time_ms_dataset_builders: int,
):
  """Call `tfds_import` on registered loggers.

  Given the number of operations which can be done at import time should be
  limited, here the registered loggers calls are deferred and done either:
   - on the following call to one of the logging method; or
   - at exit time.

  Args:
    metadata: metadata associated to import operation.
    import_time_ms_tensorflow: time (ms) it took to import TF.
    import_time_ms_dataset_builders: time (ms) it took to import DatasetBuilder
      modules.
  """
  with _import_operations_lock:
    _import_operations.append(
        (metadata, import_time_ms_tensorflow, import_time_ms_dataset_builders)
    )


def builder_init() -> _Decorator:
  """Decorator to call `builder_init` method on registered loggers."""

  @wrapt.decorator
  def decorator(function, dsbuilder, args, kwargs):
    metadata = call_metadata.CallMetadata()
    _thread_id_to_builder_init_count[metadata.thread_id] += 1
    try:
      return function(*args, **kwargs)
    except Exception:
      metadata.mark_error()
      raise
    finally:
      if _thread_id_to_builder_init_count[metadata.thread_id] == 1:
        metadata.mark_end()
        data_dir = kwargs.get("data_dir")
        config = kwargs.get("config")
        if config is not None:
          config = str(config)
        version = kwargs.get("version")
        if version is not None:
          version = str(version)
        for logger in _get_registered_loggers():
          logger.builder_init(
              metadata=metadata,
              name=dsbuilder.name,
              data_dir=data_dir,
              config=config,
              version=version,
          )
      _thread_id_to_builder_init_count[metadata.thread_id] -= 1

  return decorator


class builder_info(_DsbuilderMethodDecorator):  # pylint: disable=invalid-name
  """Decorator to call `builder_info` method on registered loggers."""

  IS_PROPERTY = True

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    del args, kwargs
    logger_method()


class as_dataset(_DsbuilderMethodDecorator):  # pylint: disable=invalid-name
  """Decorator to call `as_dataset` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    logger_method(
        split=args and args[0] or kwargs.get("split"),
        batch_size=kwargs.get("batch_size"),
        shuffle_files=kwargs.get("shuffle_files"),
        read_config=kwargs.get("read_config"),
        as_supervised=kwargs.get("as_supervised"),
        decoders=kwargs.get("decoders"),
    )


class download_and_prepare(_DsbuilderMethodDecorator):  # pylint: disable=invalid-name
  """Decorator to call `download_and_prepare` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    del args
    logger_method(
        download_dir=kwargs.get("download_dir"),
        download_config=kwargs.get("download_config"),
        file_format=kwargs.get("file_format"),
    )


def as_numpy(function: _Decorator) -> _Decorator:
  """Decorator to call `as_numpy` method on registered loggers."""

  # Here we use `functools.wraps` as `wrapt.decorator` is not serializable.
  # https://github.com/GrahamDumpleton/wrapt/issues/158.
  @functools.wraps(function)
  def decorator(*args, **kwargs):
    dataset = args[0] if args else kwargs["dataset"]
    metadata = call_metadata.CallMetadata()
    try:
      return function(*args, **kwargs)
    except Exception:
      metadata.mark_error()
      raise
    finally:
      metadata.mark_end()
      for logger in _get_registered_loggers():
        logger.as_numpy(metadata=metadata, dataset=dataset)

  return decorator


class list_builders(_FunctionDecorator):  # pylint: disable=invalid-name
  """Decorator to call `list_builders` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    del args
    logger_method(with_community_datasets=kwargs.get("with_community_datasets"))


class load_from_ml_catalog(_FunctionDecorator):  # pylint: disable=invalid-name
  """Decorator to call `load_from_ml_catalog` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    logger_method(
        name=args[0] if args else kwargs["name"],
        split=kwargs.get("split"),
        decoders=kwargs.get("decoders"),
        read_config=kwargs.get("read_config"),
        shuffle=kwargs.get("shuffle"),
    )


class load(_FunctionDecorator):  # pylint: disable=invalid-name
  """Decorator to call `load` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    logger_method(
        name=args[0] if args else kwargs["name"],
        split=kwargs.get("split"),
        data_dir=kwargs.get("data_dir"),
        batch_size=kwargs.get("batch_size"),
        shuffle_files=kwargs.get("shuffle_files"),
        download=kwargs.get("download"),
        as_supervised=kwargs.get("as_supervised"),
        decoders=kwargs.get("decoders"),
        read_config=kwargs.get("read_config"),
        with_info=kwargs.get("with_info"),
        try_gcs=kwargs.get("try_gcs"),
    )


class builder(_FunctionDecorator):  # pylint: disable=invalid-name
  """Decorator to call `builder` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    logger_method(
        name=args[0] if args else kwargs["name"],
        try_gcs=kwargs.get("try_gcs"),
    )


class dataset_collection(_FunctionDecorator):  # pylint: disable=invalid-name
  """Decorator to call `dataset_collection` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    logger_method(
        name=args[0] if args else kwargs["name"],
        loader_kwargs=kwargs.get("loader_kwargs"),
    )


class data_source(_FunctionDecorator):  # pylint: disable=invalid-name
  """Decorator to call `data_source` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    logger_method(
        name=args[0] if args else kwargs["name"],
        split=kwargs.get("split"),
        data_dir=kwargs.get("data_dir"),
        download=kwargs.get("download"),
        decoders=kwargs.get("decoders"),
        try_gcs=kwargs.get("try_gcs"),
    )


class as_data_source(_DsbuilderMethodDecorator):  # pylint: disable=invalid-name
  """Decorator to call `as_data_source` method on registered loggers."""

  def _call_logger_method(
      self,
      logger_method: _LoggerMethod,
      args: Any,
      kwargs: Any,
  ):
    logger_method(
        split=args and args[0] or kwargs.get("split"),
        decoders=kwargs.get("decoders"),
    )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""HellaSwag dataset."""

from __future__ import annotations

import json
import os

from etils import epath
import numpy as np
import tensorflow_datasets.public_api as tfds

_CITATION = """
@inproceedings{zellers2019hellaswag,
    title={HellaSwag: Can a Machine Really Finish Your Sentence?},
    author={Zellers, Rowan and Holtzman, Ari and Bisk, Yonatan and Farhadi, Ali and Choi, Yejin},
    booktitle ={Proceedings of the 57th Annual Meeting of the Association for Computational Linguistics},
    year={2019}
}
"""

_DESCRIPTION = """
The HellaSwag dataset is a benchmark for Commonsense NLI. It includes a context
and some endings which complete the context.
"""

_HELLASWAG_URL = (
    'https://raw.githubusercontent.com/rowanz/hellaswag/master/data/'
)


class Hellaswag(tfds.core.GeneratorBasedBuilder):
  """HellaSwag Dataset."""

  VERSION = tfds.core.Version('1.1.0')
  RELEASE_NOTES = {
      '1.1.0': 'Another split dimension for source (wikihow vs activitynet)',
      '1.0.0': (
          'Adding separate splits for in-domain and out-of-domain '
          'validation/test sets.'
      ),
  }
  SUPPORTED_VERSIONS = [
      tfds.core.Version('1.1.0'),
      tfds.core.Version('1.0.0'),
      tfds.core.Version('0.0.1'),
  ]

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            'context': tfds.features.Text(),
            'endings': tfds.features.Sequence(tfds.features.Text()),
            'activity_label': tfds.features.Text(),
            'label': np.int32,
            'split_type': tfds.features.Text(),
            'source_id': tfds.features.Text(),
        }),
        supervised_keys=None,
        homepage='https://rowanzellers.com/hellaswag/',
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""

    files = dl_manager.download_and_extract({
        'train': os.path.join(_HELLASWAG_URL, 'hellaswag_train.jsonl'),
        'validation': os.path.join(_HELLASWAG_URL, 'hellaswag_val.jsonl'),
        'test': os.path.join(_HELLASWAG_URL, 'hellaswag_test.jsonl'),
    })

    return {
        'train': self._generate_examples(files['train']),
        'train_activitynet': self._generate_examples(
            files['train'], source='activitynet'
        ),
        'train_wikihow': self._generate_examples(
            files['train'], source='wikihow'
        ),
        'validation': self._generate_examples(files['validation']),
        'test': self._generate_examples(files['test']),
        'validation_ind_activitynet': self._generate_examples(
            files['validation'], 'IND', 'activitynet'
        ),
        'validation_ood_activitynet': self._generate_examples(
            files['validation'], 'OOD', 'activitynet'
        ),
        'test_ind_activitynet': self._generate_examples(
            files['test'], 'IND', 'activitynet'
        ),
        'test_ood_activitynet': self._generate_examples(
            files['test'], 'OOD', 'activitynet'
        ),
        'validation_ind_wikihow': self._generate_examples(
            files['validation'], 'IND', 'wikihow'
        ),
        'validation_ood_wikihow': self._generate_examples(
            files['validation'], 'OOD', 'wikihow'
        ),
        'test_ind_wikihow': self._generate_examples(
            files['test'], 'IND', 'wikihow'
        ),
        'test_ood_wikihow': self._generate_examples(
            files['test'], 'OOD', 'wikihow'
        ),
    }

  def _generate_examples(self, filepath, domain=None, source=None):
    """Yields examples."""
    with epath.Path(filepath).open() as f:
      for idx, line in enumerate(f):
        elem = json.loads(line)
        elem_id = '%s_%d' % (os.path.basename(filepath), idx)

        if domain == 'IND' and elem['split_type'] != 'indomain':
          continue
        if domain == 'OOD' and elem['split_type'] != 'zeroshot':
          continue
        if source and not elem['source_id'].startswith(source):
          continue

        yield elem_id, {
            'context': elem['ctx'],
            'endings': elem['endings'],
            'activity_label': elem['activity_label'],
            'label': elem.get('label', -1),
            'split_type': elem['split_type'],
            'source_id': elem['source_id'],
        }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Generate XTREME-S audio test data.

This does not create the TSV files.

"""

import os

from absl import app
from absl import flags
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.testing import fake_data_utils

_TFDS_DIR = flags.DEFINE_string(
    "tfds_dir",
    str(utils.tfds_write_path()),
    "Path to tensorflow_datasets directory",
)


def _output_dir():
  return os.path.join(
      _TFDS_DIR.value, "audio", "xtreme_s", "dummy_data", "fleurs"
  )


def _filepath(language: str, split: str, wav_name: str) -> str:
  return os.path.join(
      _output_dir(), language, "audio", split, f"{wav_name}.wav"
  )


def _make_wav_file(filepath: str):
  wav_file = fake_data_utils.get_random_wav_c1(duration=2, sample=16000)
  dirname = os.path.dirname(filepath)
  if not tf.io.gfile.exists(dirname):
    tf.io.gfile.makedirs(dirname)
  tf.io.gfile.copy(wav_file, filepath, overwrite=True)


def _generate_data():
  """Generate data examples."""

  # Several examples for Afrikaans.
  language = "af_za"
  for wav_name in ["train_1", "train_2", "train_3"]:
    filepath = _filepath(language=language, split="train", wav_name=wav_name)
    _make_wav_file(filepath)

  for wav_name in ["dev_1"]:
    filepath = _filepath(language=language, split="dev", wav_name=wav_name)
    _make_wav_file(filepath)

  for wav_name in ["test_1", "test_2"]:
    filepath = _filepath(language=language, split="test", wav_name=wav_name)
    _make_wav_file(filepath)


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")
  _generate_data()


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing utilities."""

import importlib
import typing
from typing import Any, List

from tensorflow_datasets.core import registered

# pylint: disable=g-import-not-at-top,g-importing-member
if typing.TYPE_CHECKING:
  # We import testing namespace but without registering the tests datasets
  # (e.g. DummyMnist,...).
  # LINT.IfChange(pydeps)
  from tensorflow_datasets.testing.dataset_builder_testing import DatasetBuilderTestCase
  from tensorflow_datasets.testing.feature_test_case import FeatureExpectationItem
  from tensorflow_datasets.testing.feature_test_case import FeatureExpectationsTestCase
  from tensorflow_datasets.testing.feature_test_case import RaggedConstant
  from tensorflow_datasets.testing.feature_test_case import SubTestCase
  from tensorflow_datasets.testing.mocking import mock_data
  from tensorflow_datasets.testing.mocking import MockPolicy
  from tensorflow_datasets.testing.mocking import PickableDataSourceMock
  from tensorflow_datasets.testing.test_case import TestCase
  from tensorflow_datasets.testing.test_utils import assert_features_equal
  from tensorflow_datasets.testing.test_utils import DummyBeamDataset
  from tensorflow_datasets.testing.test_utils import DummyDataset
  from tensorflow_datasets.testing.test_utils import DummyDatasetCollection
  from tensorflow_datasets.testing.test_utils import DummyDatasetSharedGenerator
  from tensorflow_datasets.testing.test_utils import DummyMnist
  from tensorflow_datasets.testing.test_utils import DummyParser
  from tensorflow_datasets.testing.test_utils import DummySerializer
  from tensorflow_datasets.testing.test_utils import fake_examples_dir
  from tensorflow_datasets.testing.test_utils import make_tmp_dir
  from tensorflow_datasets.testing.test_utils import mock_kaggle_api
  from tensorflow_datasets.testing.test_utils import MockFs
  from tensorflow_datasets.testing.test_utils import rm_tmp_dir
  from tensorflow_datasets.testing.test_utils import run_in_graph_and_eager_modes
  from tensorflow_datasets.testing.test_utils import test_main
  from tensorflow_datasets.testing.test_utils import tmp_dir
  # LINT.ThenChange(:deps)
# pylint: enable=g-import-not-at-top,g-importing-member

_API = {
    # LINT.IfChange(deps)
    "assert_features_equal": "tensorflow_datasets.testing.test_utils",
    "DatasetBuilderTestCase": (
        "tensorflow_datasets.testing.dataset_builder_testing"
    ),
    "DummyBeamDataset": "tensorflow_datasets.testing.test_utils",
    "DummyDataset": "tensorflow_datasets.testing.test_utils",
    "DummyDatasetCollection": "tensorflow_datasets.testing.test_utils",
    "DummyDatasetSharedGenerator": "tensorflow_datasets.testing.test_utils",
    "DummyMnist": "tensorflow_datasets.testing.test_utils",
    "DummyParser": "tensorflow_datasets.testing.test_utils",
    "DummySerializer": "tensorflow_datasets.testing.test_utils",
    "fake_examples_dir": "tensorflow_datasets.testing.test_utils",
    "FeatureExpectationItem": "tensorflow_datasets.testing.feature_test_case",
    "FeatureExpectationsTestCase": (
        "tensorflow_datasets.testing.feature_test_case"
    ),
    # TODO(afrozm): rm from here and add as methods to TestCase
    "make_tmp_dir": "tensorflow_datasets.testing.test_utils",
    "mock_data": "tensorflow_datasets.testing.mocking",
    "mock_kaggle_api": "tensorflow_datasets.testing.test_utils",
    "MockFs": "tensorflow_datasets.testing.test_utils",
    "MockPolicy": "tensorflow_datasets.testing.mocking",
    "PickableDataSourceMock": "tensorflow_datasets.testing.mocking",
    "RaggedConstant": "tensorflow_datasets.testing.feature_test_case",
    # TODO(afrozm): rm from here and add as methods to TestCase
    "rm_tmp_dir": "tensorflow_datasets.testing.test_utils",
    "run_in_graph_and_eager_modes": "tensorflow_datasets.testing.test_utils",
    "SubTestCase": "tensorflow_datasets.testing.feature_test_case",
    "test_main": "tensorflow_datasets.testing.test_utils",
    "TestCase": "tensorflow_datasets.testing.test_case",
    # TODO(afrozm): rm from here and add as methods to TestCase
    "tmp_dir": "tensorflow_datasets.testing.test_utils",
    # LINT.ThenChange(:pydeps)
}

__all__ = list(_API)


def __dir__() -> List[str]:  # pylint: disable=invalid-name
  return __all__


def __getattr__(name: str) -> Any:  # pylint: disable=invalid-name
  if name in _API:
    module_name = _API[name]
    with registered.skip_registration():
      module = importlib.import_module(module_name)
    return getattr(module, name)
  else:
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cassava leaf dataset with images of health and diseased leaves."""

import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """\
@misc{mwebaze2019icassava,
    title={iCassava 2019Fine-Grained Visual Categorization Challenge},
    author={Ernest Mwebaze and Timnit Gebru and Andrea Frome and Solomon Nsumba and Jeremy Tusubira},
    year={2019},
    eprint={1908.02900},
    archivePrefix={arXiv},
    primaryClass={cs.CV}
}
"""

_DESCRIPTION = """\
Cassava consists of leaf images for the cassava plant depicting healthy and
four (4) disease conditions; Cassava Mosaic Disease (CMD), Cassava Bacterial
Blight (CBB), Cassava Greem Mite (CGM) and Cassava Brown Streak Disease (CBSD).
Dataset consists of a total of 9430 labelled images.
The 9430 labelled images are split into a training set (5656), a test set(1885)
and a validation set (1889). The number of images per class are unbalanced with
the two disease classes CMD and CBSD having 72% of the images.
"""

_BASE_URL = "https://storage.googleapis.com/emcassavadata/cassavaleafdata.zip"
_LABELS = ["cbb", "cbsd", "cgm", "cmd", "healthy"]


class Cassava(tfds.core.GeneratorBasedBuilder):
  """Cassava leaf image dataset."""

  VERSION = tfds.core.Version("0.1.0")

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "image": tfds.features.Image(),
            "image/filename": tfds.features.Text(),  # test-cbb-0.jpg
            "label": tfds.features.ClassLabel(names=_LABELS),
        }),
        supervised_keys=("image", "label"),
        homepage="https://www.kaggle.com/c/cassava-disease/overview",
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    path = dl_manager.download_and_extract(_BASE_URL)
    train_path = os.path.join(path, "cassavaleafdata/train")
    test_path = os.path.join(path, "cassavaleafdata/test")
    validation_path = os.path.join(path, "cassavaleafdata/validation")

    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN,
            gen_kwargs={"datapath": train_path},
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.TEST,
            gen_kwargs={"datapath": test_path},
        ),
        tfds.core.SplitGenerator(
            name=tfds.Split.VALIDATION,
            gen_kwargs={"datapath": validation_path},
        ),
    ]

  def _generate_examples(self, datapath):
    """Yields examples of cassava leaf images and labels."""
    for label in tf.io.gfile.listdir(datapath):
      for fpath in tf.io.gfile.glob(os.path.join(datapath, label, "*.jpg")):
        fname = os.path.basename(fpath)
        record = {
            "image": fpath,
            "image/filename": fname,
            "label": label,
        }
        yield fname, record



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Util to globally control which datasets can be accessed.

Different scripts may access different datasets (community datasets,...).

By default all TFDS scripts only have access to open-source non-community
datasets. This avoid accidentally accessing datasets (e.g. documenting
community datasets in the TFDS documentation).

Scripts which require additional datasets visibility should explicitly opt-in
for the requested category:

```python
tfds.core.visibility.set_availables([
    tfds.core.visibility.DatasetType.TFDS_PUBLIC,
    tfds.core.visibility.DatasetType.COMMUNITY_PUBLIC,
])
```

Updating the visibility affect the following functions:

* `tfds.load`
* `tfds.list_builders`
* `tfds.core.load.list_full_names`

"""

import contextlib
import enum
import pathlib
from typing import Iterable, Iterator, List

from absl import app


class DatasetType(enum.Enum):
  TFDS_PUBLIC = enum.auto()
  COMMUNITY_PUBLIC = enum.auto()

  def is_available(self) -> bool:
    return self in _current_available


# Default datasets availables.
# For TFDS internal scripts, this is overwritten in `_set_default_visibility`
_current_available = {
    DatasetType.TFDS_PUBLIC,
    DatasetType.COMMUNITY_PUBLIC,
}


def set_availables(new_ds_types: Iterable[DatasetType]) -> None:
  """Overwrites the current visibility permissions.

  Calling this function will affect the returned value of
  `tfds.list_builders`,...

  Args:
    new_ds_types: New ds types.
  """
  _current_available.clear()
  _current_available.update(new_ds_types)


def get_availables() -> List[DatasetType]:
  """Returns availables dataset types."""
  return sorted(_current_available, key=lambda ds_type: ds_type.name)


@contextlib.contextmanager
def set_availables_tmp(new_ds_types: Iterable[DatasetType]) -> Iterator[None]:
  """Contextmanager/decorator version of `set_availables`."""
  old_ds_types = set(_current_available)
  try:
    set_availables(new_ds_types)
    yield
  finally:
    set_availables(old_ds_types)  # Restore previous permissions


def _set_default_visibility() -> None:
  """Overwrites the default visibility for the TFDS scripts.

  If the script executed is a TFDS script, then it restricts the visibility
  to only open-source non-community datasets.
  """
  import __main__  # pytype: disable=import-error  # pylint: disable=g-import-not-at-top

  main_file = getattr(__main__, '__file__', None)
  if main_file and 'tensorflow_datasets' in pathlib.Path(main_file).parts:
    # If the script is launched from within a TFDS script, we disable community
    # datasets and restrict scripts to only public datasets.
    # Accessing community datasets should be explicitly requested.
    set_availables(
        [
            DatasetType.TFDS_PUBLIC,
        ]
    )


app.call_after_init(_set_default_visibility)
